import tensorflow as tf


def project_biases(tf_features, n_features):
    """
    Projects the biases from the feature space to calculate bias per actor
    :param tf_features:
    :param n_features:
    :return:
    """
    tf_feature_biases = tf.Variable(tf.zeros([n_features, 1]))

    # The reduce sum is to perform a rank reduction
    tf.compat.v1.disable_eager_execution()
    tf_projected_biases = tf.reduce_sum(
        tf.compat.v1.sparse_tensor_dense_matmul(tf_features, tf_feature_biases),
        axis=1
    )

    return tf_feature_biases, tf_projected_biases


def split_sparse_tensor_indices(tf_sparse_tensor, n_dimensions):
    """
    Separates the each dimension of a sparse tensor's indices into independent tensors
    :param tf_sparse_tensor:
    :param n_dimensions:
    :return:
    """
    tf_transposed_indices = tf.transpose(tf_sparse_tensor.indices)
    return (tf_transposed_indices[i] for i in range(n_dimensions))


def bias_prediction_dense(tf_prediction, tf_projected_user_biases, tf_projected_item_biases):
    """
    Broadcasts user and item biases across their respective axes of the dense predictions
    :param tf_prediction:
    :param tf_projected_user_biases:
    :param tf_projected_item_biases:
    :return:
    """
    return tf_prediction + tf.expand_dims(tf_projected_user_biases, 1) + tf.expand_dims(tf_projected_item_biases, 0)


def bias_prediction_serial(tf_prediction_serial, tf_projected_user_biases, tf_projected_item_biases, tf_x_user,
                           tf_x_item):
    """
    Calculates the bias per user/item pair and adds it to the serial predictions
    :param tf_prediction_serial:
    :param tf_projected_user_biases:
    :param tf_projected_item_biases:
    :param tf_x_user:
    :param tf_x_item:
    :return:
    """
    gathered_user_biases = tf.gather(tf_projected_user_biases, tf_x_user)
    gathered_item_biases = tf.gather(tf_projected_item_biases, tf_x_item)
    return tf_prediction_serial + gathered_user_biases + gathered_item_biases


def densify_sampled_item_predictions(tf_sample_predictions_serial, tf_n_sampled_items, tf_n_users):
    """
    Turns the serial predictions of the sample items in to a dense matrix of shape [ n_users, n_sampled_items ]
    :param tf_sample_predictions_serial:
    :param tf_n_sampled_items:
    :param tf_n_users:
    :return:
    """
    tf.compat.v1.disable_eager_execution()
    densified_shape = tf.cast(tf.compat.v1.stack([tf_n_users, tf_n_sampled_items]), tf.int32)
    densified_predictions = tf.reshape(tf_sample_predictions_serial, shape=densified_shape)
    return densified_predictions


def rank_predictions(tf_prediction):
    """
    Double-sortation serves as a ranking process
    The +1 is so the top-ranked has a non-zero rank
    :param tf_prediction:
    :return:
    """
    tf_prediction_item_size = tf.shape(tf_prediction)[1]
    tf_indices_of_ranks = tf.nn.top_k(tf_prediction, k=tf_prediction_item_size)[1]
    return tf.nn.top_k(-tf_indices_of_ranks, k=tf_prediction_item_size)[1] + 1


def collapse_mixture_of_tastes(tastes_predictions, tastes_attentions):
    """
    Collapses a list of prediction nodes in to a single prediction node.
    :param tastes_predictions:
    :param tastes_attentions:
    :return:
    """
    stacked_predictions = tf.stack(tastes_predictions)

    # If there is attention, the attentions are used to weight each prediction
    if tastes_attentions is not None:

        # Stack the attentions and perform softmax across the tastes
        tf.compat.v1.disable_eager_execution()
        stacked_attentions = tf.compat.v1.stack(tastes_attentions)
        softmax_attentions = tf.nn.softmax(stacked_attentions, axis=0)

        # The softmax'd attentions serve as weights for the taste predictiones
        weighted_predictions = tf.multiply(stacked_predictions, softmax_attentions)
        result_prediction = tf.reduce_sum(weighted_predictions, axis=0)

    # If there is no attention, the max prediction is returned
    else:
        result_prediction = tf.reduce_max(stacked_predictions, axis=0)

    return result_prediction


def relative_cosine(tf_tensor_1, tf_tensor_2):
    """
    Returns the cosine of every row in tensor_1 against every row in tensor_2.
    :param tf_tensor_1:
    :param tf_tensor_2:
    :return:
    """
    normalized_t1 = tf.nn.l2_normalize(tf_tensor_1, 1)
    normalized_t2 = tf.nn.l2_normalize(tf_tensor_2, 1)
    return tf.matmul(normalized_t1, normalized_t2, transpose_b=True)


def predict_similar_items(prediction_graph_factory, tf_item_representation, tf_similar_items_ids):
    """
    Calculates the similarity between the given item ids and all other items using the prediction graph.
    :param prediction_graph_factory:
    :param tf_item_representation:
    :param tf_similar_items_ids:
    :return:
    """
    gathered_items = tf.gather(tf_item_representation, tf_similar_items_ids)
    sims = prediction_graph_factory.connect_dense_prediction_graph(
        tf_user_representation=gathered_items,
        tf_item_representation=tf_item_representation
    )
    return sims



import tensorflow as tf

_session = None


def get_session():
    global _session

    # Build/retrieve the session if it doesn't exist
    if _session is None:
        # if tf.get_default_session() is not None:
        #     _session = tf.get_default_session()
        tf.compat.v1.disable_eager_execution()

        if tf.compat.v1.get_default_session() is not None:
            _session = tf.compat.v1.get_default_session()
        else:
            _session = tf.compat.v1.Session()

    return _session


def set_session(session):
    global _session
    _session = session



import abc
import tensorflow as tf

from .recommendation_graphs import relative_cosine


class AbstractPredictionGraph(object):
    __metaclass__ = abc.ABCMeta

    @abc.abstractmethod
    def connect_dense_prediction_graph(self, tf_user_representation, tf_item_representation):
        """
        This method is responsible for consuming user and item representations and calculating prediction scores for all
        possible user-item pairs based on these representations.
        :param tf_user_representation: tf.Tensor
        The user representations as a Tensor of shape [n_users, n_components]
        :param tf_item_representation: tf.Tensor
        The item representations as a Tensor of shape [n_items, n_components]
        :return: tf.Tensor
        The predictions as a Tensor of shape [n_users, n_items]
        """
        pass

    @abc.abstractmethod
    def connect_serial_prediction_graph(self, tf_user_representation, tf_item_representation, tf_x_user, tf_x_item):
        """
        This method is responsible for consuming user and item representations and indices and calculating prediction
        scores for particular user-item pairs.
        :param tf_user_representation: tf.Tensor
        The user representations as a Tensor of shape [n_users, n_components]
        :param tf_item_representation: tf.Tensor
        The item representations as a Tensor of shape [n_items, n_components]
        :param tf_x_user: tf.Tensor
        The users for whom to predict as a Tensor of shape [n_interactions]
        :param tf_x_item: tf.Tensor
        The items for which to predict as a Tensor of shape [n_interactions]
        :return: tf.Tensor
        The predictions as a Tensor of shape [n_interactions]
        """
        pass


class DotProductPredictionGraph(AbstractPredictionGraph):
    """
    This prediction function calculates the prediction as the dot product between the user and item representations.
    Prediction = user_repr * item_repr
    """

    def connect_dense_prediction_graph(self, tf_user_representation, tf_item_representation):
        return tf.matmul(tf_user_representation, tf_item_representation, transpose_b=True)

    def connect_serial_prediction_graph(self, tf_user_representation, tf_item_representation, tf_x_user, tf_x_item):
        gathered_user_reprs = tf.gather(tf_user_representation, tf_x_user)
        gathered_item_reprs = tf.gather(tf_item_representation, tf_x_item)
        return tf.reduce_sum(tf.multiply(gathered_user_reprs, gathered_item_reprs), axis=1)


class CosineSimilarityPredictionGraph(AbstractPredictionGraph):
    """
    This prediction function calculates the prediction as the cosine between the user and item representations.
    Prediction = cos(user_repr, item_repr)
    """

    def connect_dense_prediction_graph(self, tf_user_representation, tf_item_representation):
        return relative_cosine(tf_tensor_1=tf_user_representation, tf_tensor_2=tf_item_representation)

    def connect_serial_prediction_graph(self, tf_user_representation, tf_item_representation, tf_x_user, tf_x_item):
        normalized_users = tf.nn.l2_normalize(tf_user_representation, 1)
        normalized_items = tf.nn.l2_normalize(tf_item_representation, 1)
        gathered_user_reprs = tf.gather(normalized_users, tf_x_user)
        gathered_item_reprs = tf.gather(normalized_items, tf_x_item)
        return tf.reduce_sum(tf.multiply(gathered_user_reprs, gathered_item_reprs), axis=1)


class EuclideanSimilarityPredictionGraph(AbstractPredictionGraph):
    """
    This prediction function calculates the prediction as the negative euclidean distance between the user and
    item representations.
    Prediction = -1 * sqrt(sum((user_repr - item_repr)^2))
    """

    epsilon = 1e-16

    def connect_dense_prediction_graph(self, tf_user_representation, tf_item_representation):

        # [ n_users, 1 ]
        r_user = tf.reduce_sum(tf_user_representation ** 2, 1, keep_dims=True)

        # [ n_items, 1 ]
        r_item = tf.reduce_sum(tf_item_representation ** 2, 1, keep_dims=True)

        # [ n_users, n_items ]
        distance = (r_user
                    - 2.0 * tf.matmul(tf_user_representation, tf_item_representation, transpose_b=True)
                    + tf.transpose(r_item))

        # For numeric stability
        distance = tf.maximum(distance, self.epsilon)

        return -1.0 * tf.sqrt(distance)

    def connect_serial_prediction_graph(self, tf_user_representation, tf_item_representation, tf_x_user, tf_x_item):

        # [ n_interactions, n_components ]
        gathered_user_reprs = tf.gather(tf_user_representation, tf_x_user)
        gathered_item_reprs = tf.gather(tf_item_representation, tf_x_item)

        # [ n_interactions, n_components ]
        delta = tf.pow(gathered_user_reprs - gathered_item_reprs, 2)

        # [ n_interactions, 1 ]
        distance = tf.reduce_sum(delta, axis=1)

        # For numeric stability
        distance = tf.maximum(distance, self.epsilon)

        return -1.0 * tf.sqrt(distance)



import numpy as np
from scipy import sparse as sp
import tensorflow as tf

from .session_management import get_session


def create_tensorrec_iterator(name):
    """
    Creates a TensorFlow Iterator that is ready for the standard TensorRec data format.
    :param name: str
    The name for this Iterator.
    :return: tf.data.Iterator
    """
    tf.compat.v1.disable_eager_execution()

    return tf.compat.v1.data.Iterator.from_structure(
            output_types=(tf.int64, tf.int64, tf.float32, tf.int64, tf.int64),
            output_shapes=([None], [None], [None], [], []),
            shared_name=name
    )


def create_tensorrec_dataset_from_sparse_matrix(sparse_matrix):
    """
    Creates a TensorFlow Dataset containing the data from the given sparse matrix.
    :param sparse_matrix: scipy.sparse matrix
    The data to be contained in this Dataset.
    :return: tf.data.Dataset
    """
    if not isinstance(sparse_matrix, sp.coo_matrix):
        sparse_matrix = sp.coo_matrix(sparse_matrix)

    row_index = np.array([sparse_matrix.row], dtype=np.int64)
    col_index = np.array([sparse_matrix.col], dtype=np.int64)
    values = np.array([sparse_matrix.data], dtype=np.float32)
    n_dim_0 = np.array([sparse_matrix.shape[0]], dtype=np.int64)
    n_dim_1 = np.array([sparse_matrix.shape[1]], dtype=np.int64)

    tensor_slices = (row_index, col_index, values, n_dim_0, n_dim_1)

    return tf.data.Dataset.from_tensor_slices(tensor_slices)


def write_tfrecord_from_sparse_matrix(tfrecord_path, sparse_matrix):
    """
    Writes the contents of a sparse matrix to a TFRecord file.
    :param tfrecord_path: str
    :param sparse_matrix: scipy.sparse matrix
    :return: str
    The tfrecord path
    """
    dataset = create_tensorrec_dataset_from_sparse_matrix(sparse_matrix=sparse_matrix)
    return write_tfrecord_from_tensorrec_dataset(tfrecord_path=tfrecord_path,
                                                 dataset=dataset)


def get_dimensions_from_tensorrec_dataset(dataset):
    """
    Given a TensorFlow Dataset in the standard TensorRec format, returns the dimensions of the SparseTensor to be
    populated by the Dataset.
    :param dataset: tf.data.Dataset
    :return: (int, int)
    """
    session = get_session()
    iterator = create_tensorrec_iterator('dims_iterator')
    initializer = iterator.make_initializer(dataset)
    _, _, _, tf_d0, tf_d1 = iterator.get_next()
    session.run(initializer)
    d0, d1 = session.run([tf_d0, tf_d1])
    return d0, d1


def write_tfrecord_from_tensorrec_dataset(tfrecord_path, dataset):
    """
    Writes the contents of a TensorRec Dataset to a TFRecord file.
    :param tfrecord_path: str
    :param dataset: tf.data.Dataset
    :return: str
    The tfrecord path
    """
    session = get_session()
    iterator = create_tensorrec_iterator('dataset_writing_iterator')
    initializer = iterator.make_initializer(dataset)
    tf_row_index, tf_col_index, tf_values, tf_d0, tf_d1 = iterator.get_next()
    session.run(initializer)
    row_index, col_index, values, d0, d1 = session.run([tf_row_index, tf_col_index, tf_values, tf_d0, tf_d1])

    def _int64_feature(int_values):
        return tf.train.Feature(int64_list=tf.train.Int64List(value=int_values))

    def _float_feature(float_values):
        return tf.train.Feature(float_list=tf.train.FloatList(value=float_values))

    writer = tf.python_io.TFRecordWriter(tfrecord_path)
    feature = {
        'row_index': _int64_feature(row_index),
        'col_index': _int64_feature(col_index),
        'values': _float_feature(values),
        'd0': _int64_feature([d0]),
        'd1': _int64_feature([d1]),
    }
    example = tf.train.Example(features=tf.train.Features(feature=feature))
    writer.write(example.SerializeToString())
    writer.close()
    return tfrecord_path


def create_tensorrec_dataset_from_tfrecord(tfrecord_path):
    """
    Loads a TFRecord file and creates a Dataset with the contents.
    :param tfrecord_path: str
    :return: tf.data.Dataset
    """

    def parse_tensorrec_tfrecord(example_proto):
        
        tf.compat.v1.disable_eager_execution()

        features = {
            'row_index': tf.compat.v1.FixedLenSequenceFeature((), tf.int64, allow_missing=True),
            'col_index': tf.compat.v1.FixedLenSequenceFeature((), tf.int64, allow_missing=True),
            'values': tf.compat.v1.FixedLenSequenceFeature((), tf.float32, allow_missing=True),
            'd0': tf.compat.v1.FixedLenFeature((), tf.int64),
            'd1': tf.compat.v1.FixedLenFeature((), tf.int64),
        }
        parsed_features = tf.compat.v1.parse_single_example(example_proto, features)
        return (parsed_features['row_index'], parsed_features['col_index'], parsed_features['values'],
                parsed_features['d0'], parsed_features['d1'])

    dataset = tf.data.TFRecordDataset(tfrecord_path).map(parse_tensorrec_tfrecord)
    return dataset



from unittest import TestCase

from tensorrec import TensorRec
from tensorrec.loss_graphs import (
    RMSELossGraph, RMSEDenseLossGraph, WMRBLossGraph, BalancedWMRBLossGraph
)
from tensorrec.util import generate_dummy_data_with_indicator


class LossGraphsTestCase(TestCase):

    @classmethod
    def setUpClass(cls):
        cls.interactions, cls.user_features, cls.item_features = generate_dummy_data_with_indicator(
            num_users=10, num_items=12, interaction_density=.5)

    def test_rmse_loss(self):
        model = TensorRec(loss_graph=RMSELossGraph())
        model.fit(self.interactions, self.user_features, self.item_features, epochs=5)

    def test_rmse_loss_biased(self):
        model = TensorRec(loss_graph=RMSELossGraph(), biased=True)
        model.fit(self.interactions, self.user_features, self.item_features, epochs=5)

    def test_rmse_dense_loss(self):
        model = TensorRec(loss_graph=RMSEDenseLossGraph())
        model.fit(self.interactions, self.user_features, self.item_features, epochs=5)

    def test_rmse_dense_loss_biased(self):
        model = TensorRec(loss_graph=RMSEDenseLossGraph(), biased=True)
        model.fit(self.interactions, self.user_features, self.item_features, epochs=5)

    def test_wmrb_loss(self):
        model = TensorRec(loss_graph=WMRBLossGraph())
        model.fit(self.interactions, self.user_features, self.item_features, epochs=5, n_sampled_items=10)

    def test_wmrb_loss_biased(self):
        model = TensorRec(loss_graph=WMRBLossGraph(), biased=True)
        model.fit(self.interactions, self.user_features, self.item_features, epochs=5, n_sampled_items=10)

    def test_balanced_wmrb_loss(self):
        model = TensorRec(loss_graph=BalancedWMRBLossGraph())
        model.fit(self.interactions, self.user_features, self.item_features, epochs=5, n_sampled_items=10)

    def test_balanced_wmrb_loss_biased(self):
        model = TensorRec(loss_graph=BalancedWMRBLossGraph(), biased=True)
        model.fit(self.interactions, self.user_features, self.item_features, epochs=5, n_sampled_items=10)



from functools import partial
from itertools import cycle
import logging
import numpy as np
import os
import pickle
from scipy import sparse as sp
import tensorflow as tf

from .errors import (
    ModelNotBiasedException, ModelNotFitException, ModelWithoutAttentionException, BatchNonSparseInputException,
    TfVersionException
)
from .input_utils import create_tensorrec_iterator, get_dimensions_from_tensorrec_dataset
from .loss_graphs import AbstractLossGraph, RMSELossGraph
from .prediction_graphs import AbstractPredictionGraph, DotProductPredictionGraph
from .recommendation_graphs import (
    project_biases, split_sparse_tensor_indices, bias_prediction_dense, bias_prediction_serial, rank_predictions,
    densify_sampled_item_predictions, collapse_mixture_of_tastes, predict_similar_items
)
from .representation_graphs import AbstractRepresentationGraph, LinearRepresentationGraph
from .session_management import get_session
from .util import sample_items, calculate_batched_alpha, datasets_from_raw_input


class TensorRec(object):

    def __init__(self,
                 n_components=100,
                 n_tastes=1,
                 user_repr_graph=LinearRepresentationGraph(),
                 item_repr_graph=LinearRepresentationGraph(),
                 attention_graph=None,
                 prediction_graph=DotProductPredictionGraph(),
                 loss_graph=RMSELossGraph(),
                 biased=True,):
        """
        A TensorRec recommendation model.
        :param n_components: Integer
        The dimension of a single output of the representation function. Must be >= 1.
        :param n_tastes: Integer
        The number of tastes/reprs to be calculated for each user. Must be >= 1.
        :param user_repr_graph: AbstractRepresentationGraph
        An object which inherits AbstractRepresentationGraph that contains a method to calculate user representations.
        See tensorrec.representation_graphs for examples.
        :param item_repr_graph: AbstractRepresentationGraph
        An object which inherits AbstractRepresentationGraph that contains a method to calculate item representations.
        See tensorrec.representation_graphs for examples.
        :param attention_graph: AbstractRepresentationGraph or None
        Optional. An object which inherits AbstractRepresentationGraph that contains a method to calculate user
        attention. Any valid repr_graph is also a valid attention graph. If None, no attention process will be applied.
        :param prediction_graph: AbstractPredictionGraph
        An object which inherits AbstractPredictionGraph that contains a method to calculate predictions from a pair of
        user/item reprs.
        See tensorrec.prediction_graphs for examples.
        :param loss_graph: AbstractLossGraph
        An object which inherits AbstractLossGraph that contains a method to calculate the loss function.
        See tensorrec.loss_graphs for examples.
        :param biased: bool
        If True, a bias value will be calculated for every user feature and item feature.
        """

        # Check TensorFlow version
        major, minor, patch = tf.__version__.split(".")
        if int(major) < 1 or int(major) == 1 and int(minor) < 7:
            raise TfVersionException(tf_version=tf.__version__)

        # Arg Check
        if (n_components is None) or (n_tastes is None) or (user_repr_graph is None) or (item_repr_graph is None) \
                or (prediction_graph is None) or (loss_graph is None):
            raise ValueError("All arguments to TensorRec() must be non-None")
        if n_components < 1:
            raise ValueError("n_components must be >= 1")
        if n_tastes < 1:
            raise ValueError("n_tastes must be >= 1")
        if not isinstance(user_repr_graph, AbstractRepresentationGraph):
            raise ValueError("user_repr_graph must inherit AbstractRepresentationGraph")
        if not isinstance(item_repr_graph, AbstractRepresentationGraph):
            raise ValueError("item_repr_graph must inherit AbstractRepresentationGraph")
        if not isinstance(prediction_graph, AbstractPredictionGraph):
            raise ValueError("prediction_graph must inherit AbstractPredictionGraph")
        if not isinstance(loss_graph, AbstractLossGraph):
            raise ValueError("loss_graph must inherit AbstractLossGraph")
        if attention_graph is not None:
            if not isinstance(attention_graph, AbstractRepresentationGraph):
                raise ValueError("attention_graph must be None or inherit AbstractRepresentationGraph")
            if n_tastes == 1:
                raise ValueError("attention_graph must be None if n_tastes == 1")

        self.n_components = n_components
        self.n_tastes = n_tastes
        self.user_repr_graph_factory = user_repr_graph
        self.item_repr_graph_factory = item_repr_graph
        self.attention_graph_factory = attention_graph
        self.prediction_graph_factory = prediction_graph
        self.loss_graph_factory = loss_graph
        self.biased = biased

        # A list of the attr names of every graph hook attr
        self.graph_tensor_hook_attr_names = [

            # Top-level API nodes
            'tf_user_representation', 'tf_item_representation', 'tf_prediction_serial', 'tf_prediction', 'tf_rankings',
            'tf_predict_similar_items', 'tf_rank_similar_items',

            # Training nodes
            'tf_basic_loss', 'tf_weight_reg_loss', 'tf_loss',

            # Feed placeholders
            'tf_learning_rate', 'tf_alpha', 'tf_sample_indices', 'tf_n_sampled_items', 'tf_similar_items_ids',
        ]
        if self.biased:
            self.graph_tensor_hook_attr_names += ['tf_projected_user_biases', 'tf_projected_item_biases']
        if self.attention_graph_factory is not None:
            self.graph_tensor_hook_attr_names += ['tf_user_attention_representation']

        self.graph_operation_hook_attr_names = [
            # AdamOptimizer
            'tf_optimizer',
        ]
        self.graph_iterator_hook_attr_names = [
            # Input data iterators
            'tf_user_feature_iterator', 'tf_item_feature_iterator', 'tf_interaction_iterator',
        ]

        # Calling the break routine during __init__ creates all the attrs on the TensorRec object with an initial value
        # of None
        self._break_graph_hooks()

        # A map of every graph hook attr name to the node name after construction
        # Tensors and operations are stored separated because they are handled differently by TensorFlow
        self.graph_tensor_hook_node_names = {}
        self.graph_operation_hook_node_names = {}
        self.graph_iterator_hook_node_names = {}

    def _break_graph_hooks(self):
        for graph_tensor_hook_attr_name in self.graph_tensor_hook_attr_names:
            self.__setattr__(graph_tensor_hook_attr_name, None)
        for graph_operation_hook_attr_name in self.graph_operation_hook_attr_names:
            self.__setattr__(graph_operation_hook_attr_name, None)
        for graph_iterator_hook_attr_name in self.graph_iterator_hook_attr_names:
            self.__setattr__(graph_iterator_hook_attr_name, None)

    def _attach_graph_hooks(self):
        session = get_session()

        for graph_tensor_hook_attr_name in self.graph_tensor_hook_attr_names:
            graph_tensor_hook_node_name = self.graph_tensor_hook_node_names[graph_tensor_hook_attr_name]
            node = session.graph.get_tensor_by_name(name=graph_tensor_hook_node_name)
            self.__setattr__(graph_tensor_hook_attr_name, node)

        for graph_operation_hook_attr_name in self.graph_operation_hook_attr_names:
            graph_operation_hook_node_name = self.graph_operation_hook_node_names[graph_operation_hook_attr_name]
            node = session.graph.get_operation_by_name(name=graph_operation_hook_node_name)
            self.__setattr__(graph_operation_hook_attr_name, node)

        for graph_iterator_hook_attr_name in self.graph_iterator_hook_attr_names:
            iterator_resource_name, output_types, output_shapes, output_classes = \
                self.graph_iterator_hook_node_names[graph_iterator_hook_attr_name]
            iterator_resource = session.graph.get_tensor_by_name(name=iterator_resource_name)
            iterator = tf.data.Iterator(iterator_resource, None, output_types, output_shapes, output_classes)
            self.__setattr__(graph_iterator_hook_attr_name, iterator)

    def _record_graph_hook_names(self):

        # Record serializable node names/info for each graph hook
        for graph_tensor_hook_attr_name in self.graph_tensor_hook_attr_names:
            hook = self.__getattribute__(graph_tensor_hook_attr_name)
            self.graph_tensor_hook_node_names[graph_tensor_hook_attr_name] = hook.name

        for graph_operation_hook_attr_name in self.graph_operation_hook_attr_names:
            hook = self.__getattribute__(graph_operation_hook_attr_name)
            self.graph_operation_hook_node_names[graph_operation_hook_attr_name] = hook.name

        for graph_iterator_hook_attr_name in self.graph_iterator_hook_attr_names:
            hook = self.__getattribute__(graph_iterator_hook_attr_name)
            iterator_resource_name = hook._iterator_resource.name
            output_types = hook.output_types
            output_shapes = hook.output_shapes
            output_classes = hook.output_classes
            self.graph_iterator_hook_node_names[graph_iterator_hook_attr_name] = (
                iterator_resource_name, output_types, output_shapes, output_classes
            )

    def _create_batched_dataset_initializers(self, interactions, user_features, item_features, user_batch_size=None):

        if user_batch_size is not None:

            # Raise exception if interactions and user_features aren't sparse matrices
            if (not sp.issparse(interactions)) or (not sp.issparse(user_features)):
                raise BatchNonSparseInputException()

            # Coerce to CSR for fast batching
            if not isinstance(interactions, sp.csr_matrix):
                interactions = sp.csr_matrix(interactions)
            if not isinstance(user_features, sp.csr_matrix):
                user_features = sp.csr_matrix(user_features)

            n_users = user_features.shape[0]

            interactions_batched = []
            user_features_batched = []

            start_batch = 0
            while start_batch < n_users:

                # min() ensures that the batch bounds doesn't go past the end of the index
                end_batch = min(start_batch + user_batch_size, n_users)

                interactions_batched.append(interactions[start_batch:end_batch])
                user_features_batched.append(user_features[start_batch:end_batch])

                start_batch = end_batch

            # Overwrite the input with the new, batched input
            interactions = interactions_batched
            user_features = user_features_batched

        # TODO this is hand-wavy and begging for a cleaner refactor
        (int_ds, uf_ds, if_ds), (int_init, uf_init, if_init) = self._create_datasets_and_initializers(
            interactions=interactions, user_features=user_features, item_features=item_features
        )

        # Ensure that lengths make sense
        if len(int_init) != len(uf_init):
            raise ValueError('Number of batches in user_features and interactions must be equal.')
        if (len(if_init) > 1) and (len(if_init) != len(uf_init)):
            raise ValueError('Number of batches in item_features must be 1 or equal to the number of batches in '
                             'user_features.')

        # Cycle item features when zipping because there should only be one
        datasets = [ds_set for ds_set in zip(int_ds, uf_ds, cycle(if_ds))]
        initializers = [init_set for init_set in zip(int_init, uf_init, cycle(if_init))]

        return datasets, initializers

    def _create_datasets_and_initializers(self, interactions=None, user_features=None, item_features=None):

        datasets = []
        initializers = []

        if interactions is not None:
            interactions_datasets = datasets_from_raw_input(raw_input=interactions)
            interactions_initializers = [self.tf_interaction_iterator.make_initializer(dataset)
                                         for dataset in interactions_datasets]
            datasets.append(interactions_datasets)
            initializers.append(interactions_initializers)

        if user_features is not None:
            user_features_datasets = datasets_from_raw_input(raw_input=user_features)
            user_features_initializers = [self.tf_user_feature_iterator.make_initializer(dataset)
                                          for dataset in user_features_datasets]
            datasets.append(user_features_datasets)
            initializers.append(user_features_initializers)

        if item_features is not None:
            item_features_datasets = datasets_from_raw_input(raw_input=item_features)
            item_features_initializers = [self.tf_item_feature_iterator.make_initializer(dataset)
                                          for dataset in item_features_datasets]
            datasets.append(item_features_datasets)
            initializers.append(item_features_initializers)

        return datasets, initializers

    def _build_input_iterators(self):
        self.tf_user_feature_iterator = create_tensorrec_iterator(name='tf_user_feature_iterator')
        self.tf_item_feature_iterator = create_tensorrec_iterator(name='tf_item_feature_iterator')
        self.tf_interaction_iterator = create_tensorrec_iterator(name='tf_interaction_iterator')

    def _build_tf_graph(self, n_user_features, n_item_features):
            

        # Build placeholders
        tf.compat.v1.disable_eager_execution()
        self.tf_n_sampled_items = tf.compat.v1.placeholder('int64')
        self.tf_similar_items_ids = tf.compat.v1.placeholder('int64', [None])
        self.tf_learning_rate = tf.compat.v1.placeholder('float', None)
        self.tf_alpha = tf.compat.v1.placeholder('float', None)

        tf_user_feature_rows, tf_user_feature_cols, tf_user_feature_values, tf_n_users, _ = \
            self.tf_user_feature_iterator.get_next()
        tf_item_feature_rows, tf_item_feature_cols, tf_item_feature_values, tf_n_items, _ = \
            self.tf_item_feature_iterator.get_next()
        tf_interaction_rows, tf_interaction_cols, tf_interaction_values, _, _ = \
            self.tf_interaction_iterator.get_next()

        tf_user_feature_indices = tf.compat.v1.stack([tf_user_feature_rows, tf_user_feature_cols], axis=1)
        tf_item_feature_indices = tf.compat.v1.stack([tf_item_feature_rows, tf_item_feature_cols], axis=1)
        tf_interaction_indices = tf.compat.v1.stack([tf_interaction_rows, tf_interaction_cols], axis=1)

        # Construct the features and interactions as sparse matrices
        tf_user_features = tf.SparseTensor(tf_user_feature_indices, tf_user_feature_values,
                                           [tf_n_users, n_user_features])
        tf_item_features = tf.SparseTensor(tf_item_feature_indices, tf_item_feature_values,
                                           [tf_n_items, n_item_features])
        tf_interactions = tf.SparseTensor(tf_interaction_indices, tf_interaction_values,
                                          [tf_n_users, tf_n_items])

        # Construct the sampling py_func
        # tf version error solution : v1 to v2 
        tf.compat.v1.disable_eager_execution()

        sample_items_partial = partial(sample_items, replace=self.loss_graph_factory.is_sampled_with_replacement)
        self.tf_sample_indices = tf.compat.v1.py_func(func=sample_items_partial,
                                            inp=[tf_n_items, tf_n_users, self.tf_n_sampled_items],
                                            Tout=tf.int64)
        self.tf_sample_indices.set_shape([None, None])

        # Collect the weights for regularization
        tf_weights = []

        # Build the item representations
        self.tf_item_representation, item_weights = \
            self.item_repr_graph_factory.connect_representation_graph(tf_features=tf_item_features,
                                                                      n_components=self.n_components,
                                                                      n_features=n_item_features,
                                                                      node_name_ending='item')
        tf_weights.extend(item_weights)

        tf_x_user, tf_x_item = split_sparse_tensor_indices(tf_sparse_tensor=tf_interactions, n_dimensions=2)
        tf_transposed_sample_indices = tf.transpose(self.tf_sample_indices)
        tf_x_user_sample = tf_transposed_sample_indices[0]
        tf_x_item_sample = tf_transposed_sample_indices[1]

        # These lists will hold the reprs and predictions for each taste
        tastes_tf_user_representations = []
        tastes_tf_predictions = []
        tastes_tf_prediction_serials = []
        tastes_tf_sample_prediction_serials = []

        # If this model does not use attention, Nones are used as sentinels in place of the attentions
        if self.attention_graph_factory is not None:
            tastes_tf_attentions = []
            tastes_tf_attention_serials = []
            tastes_tf_sample_attention_serials = []
            tastes_tf_attention_representations = []
        else:
            tastes_tf_attentions = None
            tastes_tf_attention_serials = None
            tastes_tf_sample_attention_serials = None
            tastes_tf_attention_representations = None

        # Build n_tastes user representations and predictions
        for taste in range(self.n_tastes):
            tf_user_representation, user_weights = \
                self.user_repr_graph_factory.connect_representation_graph(tf_features=tf_user_features,
                                                                          n_components=self.n_components,
                                                                          n_features=n_user_features,
                                                                          node_name_ending='user_{}'.format(taste))
            tastes_tf_user_representations.append(tf_user_representation)
            tf_weights.extend(user_weights)

            # Connect attention, if applicable
            if self.attention_graph_factory is not None:
                tf_attention_representation, attention_weights = \
                    self.attention_graph_factory.connect_representation_graph(tf_features=tf_user_features,
                                                                              n_components=self.n_components,
                                                                              n_features=n_user_features,
                                                                              node_name_ending='attn_{}'.format(taste))
                tf_weights.extend(attention_weights)

                tf_attention = self.prediction_graph_factory.connect_dense_prediction_graph(
                    tf_user_representation=tf_attention_representation,
                    tf_item_representation=self.tf_item_representation
                )
                tf_attention_serial = self.prediction_graph_factory.connect_serial_prediction_graph(
                    tf_user_representation=tf_attention_representation,
                    tf_item_representation=self.tf_item_representation,
                    tf_x_user=tf_x_user,
                    tf_x_item=tf_x_item,
                )
                tf_sample_attention_serial = self.prediction_graph_factory.connect_serial_prediction_graph(
                    tf_user_representation=tf_user_representation,
                    tf_item_representation=self.tf_item_representation,
                    tf_x_user=tf_x_user_sample,
                    tf_x_item=tf_x_item_sample,
                )

                tastes_tf_attentions.append(tf_attention)
                tastes_tf_attention_serials.append(tf_attention_serial)
                tastes_tf_sample_attention_serials.append(tf_sample_attention_serial)
                tastes_tf_attention_representations.append(tf_attention_representation)

            # Connect the configurable prediction graphs for each taste
            tf_prediction = self.prediction_graph_factory.connect_dense_prediction_graph(
                tf_user_representation=tf_user_representation,
                tf_item_representation=self.tf_item_representation
            )
            tf_prediction_serial = self.prediction_graph_factory.connect_serial_prediction_graph(
                tf_user_representation=tf_user_representation,
                tf_item_representation=self.tf_item_representation,
                tf_x_user=tf_x_user,
                tf_x_item=tf_x_item,
            )
            tf_sample_predictions_serial = self.prediction_graph_factory.connect_serial_prediction_graph(
                tf_user_representation=tf_user_representation,
                tf_item_representation=self.tf_item_representation,
                tf_x_user=tf_x_user_sample,
                tf_x_item=tf_x_item_sample,
            )

            # Append to tastes
            tastes_tf_predictions.append(tf_prediction)
            tastes_tf_prediction_serials.append(tf_prediction_serial)
            tastes_tf_sample_prediction_serials.append(tf_sample_predictions_serial)

        # If attention is in the graph, build the API node
        tf.compat.v1.disable_eager_execution()

        if self.attention_graph_factory is not None:
            self.tf_user_attention_representation = tf.compat.v1.stack(tastes_tf_attention_representations)

        self.tf_user_representation = tf.compat.v1.stack(tastes_tf_user_representations)
        self.tf_prediction = collapse_mixture_of_tastes(
            tastes_predictions=tastes_tf_predictions,
            tastes_attentions=tastes_tf_attentions
        )
        self.tf_prediction_serial = collapse_mixture_of_tastes(
            tastes_predictions=tastes_tf_prediction_serials,
            tastes_attentions=tastes_tf_attention_serials
        )
        tf_sample_predictions_serial = collapse_mixture_of_tastes(
            tastes_predictions=tastes_tf_sample_prediction_serials,
            tastes_attentions=tastes_tf_sample_attention_serials
        )

        # Add biases, if this is a biased estimator
        if self.biased:
            tf_user_feature_biases, self.tf_projected_user_biases = project_biases(
                tf_features=tf_user_features, n_features=n_user_features
            )
            tf_item_feature_biases, self.tf_projected_item_biases = project_biases(
                tf_features=tf_item_features, n_features=n_item_features
            )

            tf_weights.append(tf_user_feature_biases)
            tf_weights.append(tf_item_feature_biases)

            self.tf_prediction = bias_prediction_dense(
                tf_prediction=self.tf_prediction,
                tf_projected_user_biases=self.tf_projected_user_biases,
                tf_projected_item_biases=self.tf_projected_item_biases)

            self.tf_prediction_serial = bias_prediction_serial(
                tf_prediction_serial=self.tf_prediction_serial,
                tf_projected_user_biases=self.tf_projected_user_biases,
                tf_projected_item_biases=self.tf_projected_item_biases,
                tf_x_user=tf_x_user,
                tf_x_item=tf_x_item)

            tf_sample_predictions_serial = bias_prediction_serial(
                tf_prediction_serial=tf_sample_predictions_serial,
                tf_projected_user_biases=self.tf_projected_user_biases,
                tf_projected_item_biases=self.tf_projected_item_biases,
                tf_x_user=tf_x_user_sample,
                tf_x_item=tf_x_item_sample)

        tf_interactions_serial = tf_interactions.values

        # Construct API nodes
        self.tf_rankings = rank_predictions(tf_prediction=self.tf_prediction)
        self.tf_predict_similar_items = predict_similar_items(prediction_graph_factory=self.prediction_graph_factory,
                                                              tf_item_representation=self.tf_item_representation,
                                                              tf_similar_items_ids=self.tf_similar_items_ids)
        self.tf_rank_similar_items = rank_predictions(tf_prediction=self.tf_predict_similar_items)

        # Compose loss function args
        # This composition is for execution safety: it prevents loss functions that are incorrectly configured from
        # having visibility of certain nodes.
        loss_graph_kwargs = {
            'tf_prediction_serial': self.tf_prediction_serial,
            'tf_interactions_serial': tf_interactions_serial,
            'tf_interactions': tf_interactions,
            'tf_n_users': tf_n_users,
            'tf_n_items': tf_n_items,
        }
        if self.loss_graph_factory.is_dense:
            loss_graph_kwargs.update({
                'tf_prediction': self.tf_prediction,
                'tf_rankings': self.tf_rankings,
            })
        if self.loss_graph_factory.is_sample_based:
            tf_sample_predictions = densify_sampled_item_predictions(
                tf_sample_predictions_serial=tf_sample_predictions_serial,
                tf_n_sampled_items=self.tf_n_sampled_items,
                tf_n_users=tf_n_users,
            )
            loss_graph_kwargs.update({'tf_sample_predictions': tf_sample_predictions,
                                      'tf_n_sampled_items': self.tf_n_sampled_items})

        # Build loss graph
        tf.compat.v1.disable_eager_execution()

        self.tf_basic_loss = self.loss_graph_factory.connect_loss_graph(**loss_graph_kwargs)

        self.tf_weight_reg_loss = sum(tf.nn.l2_loss(weights) for weights in tf_weights)
        self.tf_loss = self.tf_basic_loss + (self.tf_alpha * self.tf_weight_reg_loss)
        self.tf_optimizer = tf.compat.v1.train.AdamOptimizer(learning_rate=self.tf_learning_rate).minimize(self.tf_loss)

        # Record the new node names
        self._record_graph_hook_names()

    def fit(self, interactions, user_features, item_features, epochs=100, learning_rate=0.1, alpha=0.00001,
            verbose=False, user_batch_size=None, n_sampled_items=None):
        """
        Constructs the TensorRec graph and fits the model.
        :param interactions: scipy.sparse matrix, tensorflow.data.Dataset, str, or list
        A matrix of interactions of shape [n_users, n_items].
        If a Dataset, the Dataset must follow the format used in tensorrec.input_utils.
        If a str, the string must be the path to a TFRecord file.
        If a list, the list must contain scipy.sparse matrices, tensorflow.data.Datasets, or strs.
        :param user_features: scipy.sparse matrix, tensorflow.data.Dataset, str, or list
        A matrix of user features of shape [n_users, n_user_features].
        If a Dataset, the Dataset must follow the format used in tensorrec.input_utils.
        If a str, the string must be the path to a TFRecord file.
        If a list, the list must contain scipy.sparse matrices, tensorflow.data.Datasets, or strs.
        :param item_features: scipy.sparse matrix, tensorflow.data.Dataset, str, or list
        A matrix of item features of shape [n_items, n_item_features].
        If a Dataset, the Dataset must follow the format used in tensorrec.input_utils.
        If a str, the string must be the path to a TFRecord file.
        If a list, the list must contain scipy.sparse matrices, tensorflow.data.Datasets, or strs.
        :param epochs: Integer
        The number of epochs to fit the model.
        :param learning_rate: Float
        The learning rate of the model.
        :param alpha:
        The weight regularization loss coefficient.
        :param verbose: boolean
        If true, the model will print a number of status statements during fitting.
        :param user_batch_size: int or None
        The maximum number of users per batch, or None for all users.
        :param n_sampled_items: int or None
        The number of items to sample per user for use in loss functions. Must be non-None if
        self.loss_graph_factory.is_sample_based is True.
        """

        # Pass-through to fit_partial
        self.fit_partial(interactions=interactions,
                         user_features=user_features,
                         item_features=item_features,
                         epochs=epochs,
                         learning_rate=learning_rate,
                         alpha=alpha,
                         verbose=verbose,
                         user_batch_size=user_batch_size,
                         n_sampled_items=n_sampled_items)

    def fit_partial(self, interactions, user_features, item_features, epochs=1, learning_rate=0.1,
                    alpha=0.00001, verbose=False, user_batch_size=None, n_sampled_items=None):
        """
        Constructs the TensorRec graph and fits the model.
        :param interactions: scipy.sparse matrix, tensorflow.data.Dataset, str, or list
        A matrix of interactions of shape [n_users, n_items].
        If a Dataset, the Dataset must follow the format used in tensorrec.input_utils.
        If a str, the string must be the path to a TFRecord file.
        If a list, the list must contain scipy.sparse matrices, tensorflow.data.Datasets, or strs.
        :param user_features: scipy.sparse matrix, tensorflow.data.Dataset, str, or list
        A matrix of user features of shape [n_users, n_user_features].
        If a Dataset, the Dataset must follow the format used in tensorrec.input_utils.
        If a str, the string must be the path to a TFRecord file.
        If a list, the list must contain scipy.sparse matrices, tensorflow.data.Datasets, or strs.
        :param item_features: scipy.sparse matrix, tensorflow.data.Dataset, str, or list
        A matrix of item features of shape [n_items, n_item_features].
        If a Dataset, the Dataset must follow the format used in tensorrec.input_utils.
        If a str, the string must be the path to a TFRecord file.
        If a list, the list must contain scipy.sparse matrices, tensorflow.data.Datasets, or strs.
        :param epochs: Integer
        The number of epochs to fit the model.
        :param learning_rate: Float
        The learning rate of the model.
        :param alpha:
        The weight regularization loss coefficient.
        :param verbose: boolean
        If true, the model will print a number of status statements during fitting.
        :param user_batch_size: int or None
        The maximum number of users per batch, or None for all users.
        :param n_sampled_items: int or None
        The number of items to sample per user for use in loss functions. Must be non-None if
        self.loss_graph_factory.is_sample_based is True.
        """

        session = get_session()

        # Arg checking
        if self.loss_graph_factory.is_sample_based:
            if (n_sampled_items is None) or (n_sampled_items <= 0):
                raise ValueError("n_sampled_items must be an integer >0")
        if (n_sampled_items is not None) and (not self.loss_graph_factory.is_sample_based):
            logging.warning('n_sampled_items was specified, but the loss graph is not sample-based')

        # Check if the iterators have been constructed. If not, build them.
        if self.tf_interaction_iterator is None:
            self._build_input_iterators()

        if verbose:
            logging.info('Processing interaction and feature data')

        dataset_sets, initializer_sets = self._create_batched_dataset_initializers(interactions=interactions,
                                                                                   user_features=user_features,
                                                                                   item_features=item_features,
                                                                                   user_batch_size=user_batch_size)

        # Check if the graph has been constructed by checking the dense prediction node
        # If it hasn't been constructed, initialize it
        if self.tf_prediction is None:

            # Check input dimensions
            first_batch = dataset_sets[0]
            _, n_user_features = get_dimensions_from_tensorrec_dataset(first_batch[1])
            _, n_item_features = get_dimensions_from_tensorrec_dataset(first_batch[2])

            # Numbers of features are either learned at fit time from the shape of these two matrices or specified at
            # TensorRec construction and cannot be changed.
            tf.compat.v1.disable_eager_execution()
            self._build_tf_graph(n_user_features=n_user_features, n_item_features=n_item_features)
            session.run(tf.compat.v1.global_variables_initializer())

        # Build the shared feed dict
        feed_dict = {self.tf_learning_rate: learning_rate,
                     self.tf_alpha: calculate_batched_alpha(num_batches=len(initializer_sets), alpha=alpha)}
        if self.loss_graph_factory.is_sample_based:
            feed_dict[self.tf_n_sampled_items] = n_sampled_items

        if verbose:
            logging.info('Beginning fitting')

        for epoch in range(epochs):
            for batch, initializers in enumerate(initializer_sets):

                session.run(initializers)
                if not verbose:
                    session.run(self.tf_optimizer, feed_dict=feed_dict)

                else:
                    _, loss, serial_predictions, wr_loss = session.run(
                        [self.tf_optimizer, self.tf_basic_loss, self.tf_prediction_serial, self.tf_weight_reg_loss],
                        feed_dict=feed_dict
                    )
                    mean_loss = np.mean(loss)
                    mean_pred = np.mean(serial_predictions)
                    weight_reg_l2_loss = alpha * wr_loss
                    logging.info('EPOCH {} BATCH {} loss = {}, weight_reg_l2_loss = {}, mean_pred = {}'.format(
                        epoch, batch, mean_loss, weight_reg_l2_loss, mean_pred
                    ))

    def predict(self, user_features, item_features):
        """
        Predict recommendation scores for the given users and items.
        :param user_features: scipy.sparse matrix, tensorflow.data.Dataset, str, or list
        A matrix of user features of shape [n_users, n_user_features].
        If a Dataset, the Dataset must follow the format used in tensorrec.input_utils.
        If a str, the string must be the path to a TFRecord file.
        If a list, the list must contain scipy.sparse matrices, tensorflow.data.Datasets, or strs.
        :param item_features: scipy.sparse matrix, tensorflow.data.Dataset, str, or list
        A matrix of item features of shape [n_items, n_item_features].
        If a Dataset, the Dataset must follow the format used in tensorrec.input_utils.
        If a str, the string must be the path to a TFRecord file.
        If a list, the list must contain scipy.sparse matrices, tensorflow.data.Datasets, or strs.
        :return: np.ndarray
        The predictions in an ndarray of shape [n_users, n_items]
        """

        # Ensure that the model has been fit
        if self.tf_prediction is None:
            raise ModelNotFitException(method='predict')

        _, initializers = self._create_datasets_and_initializers(interactions=None,
                                                                 user_features=user_features,
                                                                 item_features=item_features)
        get_session().run(initializers)

        predictions = self.tf_prediction.eval(session=get_session())

        return predictions

    def predict_similar_items(self, item_features, item_ids, n_similar):
        """
        Predicts the most similar items to the given item_ids.
        :param item_features: scipy.sparse matrix, tensorflow.data.Dataset, str, or list
        A matrix of item features of shape [n_items, n_item_features].
        If a Dataset, the Dataset must follow the format used in tensorrec.input_utils.
        If a str, the string must be the path to a TFRecord file.
        If a list, the list must contain scipy.sparse matrices, tensorflow.data.Datasets, or strs.
        :param item_ids: list or np.array
        The ids of the items of interest.
        E.g. [4, 8, 12] to get sims for items 4, 8, and 12.
        :param n_similar: int
        The number of similar items to get per item of interest.
        :return: list of lists of tuples
        The first level list corresponds to input arg item_ids.
        The second level list is of length n_similar and contains tuples of (item_id, score) for each similar item.
        """

        # Ensure that the model has been fit
        if self.tf_prediction is None:
            raise ModelNotFitException(method='predict_similar_items')

        _, initializers = self._create_datasets_and_initializers(interactions=None,
                                                                 user_features=None,
                                                                 item_features=item_features)
        get_session().run(initializers)

        feed_dict = {self.tf_similar_items_ids: np.array(item_ids)}
        sims = self.tf_predict_similar_items.eval(session=get_session(), feed_dict=feed_dict)

        results = []
        for i in range(len(item_ids)):
            item_sims = sims[i]
            best = np.argpartition(item_sims, -n_similar)[-n_similar:]
            item_results = sorted(zip(best, item_sims[best]), key=lambda x: -x[1])
            results.append(item_results)

        return results

    def predict_rank(self, user_features, item_features):
        """
        Predict recommendation ranks for the given users and items.
        :param user_features: scipy.sparse matrix, tensorflow.data.Dataset, str, or list
        A matrix of user features of shape [n_users, n_user_features].
        If a Dataset, the Dataset must follow the format used in tensorrec.input_utils.
        If a str, the string must be the path to a TFRecord file.
        If a list, the list must contain scipy.sparse matrices, tensorflow.data.Datasets, or strs.
        :param item_features: scipy.sparse matrix, tensorflow.data.Dataset, str, or list
        A matrix of item features of shape [n_items, n_item_features].
        If a Dataset, the Dataset must follow the format used in tensorrec.input_utils.
        If a str, the string must be the path to a TFRecord file.
        If a list, the list must contain scipy.sparse matrices, tensorflow.data.Datasets, or strs.
        :return: np.ndarray
        The ranks in an ndarray of shape [n_users, n_items]
        """

        # Ensure that the model has been fit
        if self.tf_prediction is None:
            raise ModelNotFitException(method='predict_rank')

        _, initializers = self._create_datasets_and_initializers(interactions=None,
                                                                 user_features=user_features,
                                                                 item_features=item_features)
        get_session().run(initializers)

        rankings = self.tf_rankings.eval(session=get_session())

        return rankings

    def predict_user_representation(self, user_features):
        """
        Predict latent representation vectors for the given users.
        :param user_features: scipy.sparse matrix, tensorflow.data.Dataset, str, or list
        A matrix of user features of shape [n_users, n_user_features].
        If a Dataset, the Dataset must follow the format used in tensorrec.input_utils.
        If a str, the string must be the path to a TFRecord file.
        If a list, the list must contain scipy.sparse matrices, tensorflow.data.Datasets, or strs.
        :return: np.ndarray
        The latent user representations in an ndarray of shape [n_users, n_components]
        """

        # Ensure that the model has been fit
        if self.tf_prediction is None:
            raise ModelNotFitException(method='predict_user_representation')

        _, initializers = self._create_datasets_and_initializers(interactions=None,
                                                                 user_features=user_features,
                                                                 item_features=None)
        get_session().run(initializers)

        user_repr = self.tf_user_representation.eval(session=get_session())

        # If there is only one user repr per user, collapse from rank 3 to rank 2
        if self.n_tastes == 1:
            user_repr = np.sum(user_repr, axis=0)

        return user_repr

    def predict_user_attention_representation(self, user_features):
        """
        Predict latent attention representation vectors for the given users.
        :param user_features: scipy.sparse matrix, tensorflow.data.Dataset, str, or list
        A matrix of user features of shape [n_users, n_user_features].
        If a Dataset, the Dataset must follow the format used in tensorrec.input_utils.
        If a str, the string must be the path to a TFRecord file.
        If a list, the list must contain scipy.sparse matrices, tensorflow.data.Datasets, or strs.
        :return: np.ndarray
        The latent user attention representations in an ndarray of shape [n_users, n_components]
        """

        # Ensure that the model has been fit
        if self.tf_prediction is None:
            raise ModelNotFitException(method='predict_user_attention_representation')

        if self.attention_graph_factory is None:
            raise ModelWithoutAttentionException()

        _, initializers = self._create_datasets_and_initializers(interactions=None,
                                                                 user_features=user_features,
                                                                 item_features=None)
        get_session().run(initializers)
        user_attn_repr = self.tf_user_attention_representation.eval(session=get_session())

        # If there is only one user attn repr per user, collapse from rank 3 to rank 2
        if self.n_tastes == 1:
            user_attn_repr = np.sum(user_attn_repr, axis=0)

        return user_attn_repr

    def predict_item_representation(self, item_features):
        """
        Predict representation vectors for the given items.
        :param item_features: scipy.sparse matrix, tensorflow.data.Dataset, str, or list
        A matrix of item features of shape [n_items, n_item_features].
        If a Dataset, the Dataset must follow the format used in tensorrec.input_utils.
        If a str, the string must be the path to a TFRecord file.
        If a list, the list must contain scipy.sparse matrices, tensorflow.data.Datasets, or strs.
        :return: np.ndarray
        The latent item representations in an ndarray of shape [n_items, n_components]
        """

        # Ensure that the model has been fit
        if self.tf_prediction is None:
            raise ModelNotFitException(method='predict_item_representation')

        _, initializers = self._create_datasets_and_initializers(interactions=None,
                                                                 user_features=None,
                                                                 item_features=item_features)
        get_session().run(initializers)
        item_repr = self.tf_item_representation.eval(session=get_session())
        return item_repr

    def predict_user_bias(self, user_features):
        """
        Predict bias values for the given users.
        :param user_features: scipy.sparse matrix, tensorflow.data.Dataset, str, or list
        A matrix of user features of shape [n_users, n_user_features].
        If a Dataset, the Dataset must follow the format used in tensorrec.input_utils.
        If a str, the string must be the path to a TFRecord file.
        If a list, the list must contain scipy.sparse matrices, tensorflow.data.Datasets, or strs.
        :return: np.ndarray
        The user biases in an ndarray of shape [n_users]
        """

        # Ensure that the model has been fit
        if self.tf_prediction is None:
            raise ModelNotFitException(method='predict_user_bias')

        if not self.biased:
            raise ModelNotBiasedException(actor='user')

        _, initializers = self._create_datasets_and_initializers(interactions=None,
                                                                 user_features=user_features,
                                                                 item_features=None)
        get_session().run(initializers)
        predictions = self.tf_projected_user_biases.eval(session=get_session())
        return predictions

    def predict_item_bias(self, item_features):
        """
        Predict bias values for the given items.
        :param item_features: scipy.sparse matrix, tensorflow.data.Dataset, str, or list
        A matrix of item features of shape [n_items, n_item_features].
        If a Dataset, the Dataset must follow the format used in tensorrec.input_utils.
        If a str, the string must be the path to a TFRecord file.
        If a list, the list must contain scipy.sparse matrices, tensorflow.data.Datasets, or strs.
        :return: np.ndarray
        The item biases in an ndarray of shape [n_items]
        """

        # Ensure that the model has been fit
        if self.tf_prediction is None:
            raise ModelNotFitException(method='predict_item_bias')

        if not self.biased:
            raise ModelNotBiasedException(actor='item')

        _, initializers = self._create_datasets_and_initializers(interactions=None,
                                                                 user_features=None,
                                                                 item_features=item_features)
        get_session().run(initializers)
        predictions = self.tf_projected_item_biases.eval(session=get_session())
        return predictions

    def save_model(self, directory_path):
        """
        Saves the model to files in the given directory.
        :param directory_path: str
        The path to the directory in which to save the model.
        :return:
        """

        # Ensure that the model has been fit
        if self.tf_prediction is None:
            raise ModelNotFitException(method='save_model')

        if not os.path.exists(directory_path):
            os.makedirs(directory_path)
        tf.compat.v1.disable_eager_execution()

        saver = tf.compat.v1.train.Saver()
        session_path = os.path.join(directory_path, 'tensorrec_session.cpkt')
        saver.save(sess=get_session(), save_path=session_path)

        # Break connections to the graph before saving the python object
        self._break_graph_hooks()
        tensorrec_path = os.path.join(directory_path, 'tensorrec.pkl')
        with open(tensorrec_path, 'wb') as file:
            pickle.dump(file=file, obj=self)

        # Reconnect to the graph after saving
        self._attach_graph_hooks()

    @classmethod
    def load_model(cls, directory_path):
        """
        Loads the TensorRec model and TensorFlow session saved in the given directory.
        :param directory_path: str
        The path to the directory containing the saved model.
        :return:
        """
        tf.compat.v1.disable_eager_execution()
        graph_path = os.path.join(directory_path, 'tensorrec_session.cpkt.meta')
        saver = tf.compat.v1.train.import_meta_graph(graph_path)

        session_path = os.path.join(directory_path, 'tensorrec_session.cpkt')
        saver.restore(sess=get_session(), save_path=session_path)

        tensorrec_path = os.path.join(directory_path, 'tensorrec.pkl')
        with open(tensorrec_path, 'rb') as file:
            model = pickle.load(file=file)
        model._attach_graph_hooks()
        return model



from unittest import TestCase

from tensorrec import TensorRec

from test.datasets import get_movielens_100k


class MovieLensTestCase(TestCase):

    @classmethod
    def setUpClass(cls):
        cls.movielens_100k = get_movielens_100k()

    def test_movie_lens_fit(self):
        """
        This test checks whether the movielens getter works and that the resulting data is viable for fitting/testing a
        TensorRec model.
        """
        train_interactions, test_interactions, user_features, item_features, _ = self.movielens_100k

        model = TensorRec()
        model.fit(interactions=train_interactions,
                  user_features=user_features,
                  item_features=item_features,
                  epochs=5)
        predictions = model.predict(user_features=user_features,
                                    item_features=item_features)

        self.assertIsNotNone(predictions)



import math
import numpy as np
from unittest import TestCase

from tensorrec import TensorRec
from tensorrec.prediction_graphs import (
    DotProductPredictionGraph, CosineSimilarityPredictionGraph, EuclideanSimilarityPredictionGraph
)
from tensorrec.session_management import get_session
from tensorrec.util import generate_dummy_data_with_indicator


class PredictionGraphsTestCase(TestCase):

    @classmethod
    def setUpClass(cls):
        cls.interactions, cls.user_features, cls.item_features = generate_dummy_data_with_indicator(
            num_users=10, num_items=12, interaction_density=.5)

    def test_dot_product(self):
        model = TensorRec(prediction_graph=DotProductPredictionGraph())
        model.fit(self.interactions, self.user_features, self.item_features, epochs=5)

    def test_cos_distance(self):
        model = TensorRec(prediction_graph=CosineSimilarityPredictionGraph())
        model.fit(self.interactions, self.user_features, self.item_features, epochs=5)


class DotProductTestCase(TestCase):

    @classmethod
    def setUpClass(cls):
        cls.session = get_session()

    def test_dense_prediction(self):
        graph = DotProductPredictionGraph()
        array_1 = np.array([
            [1.0, 1.0],
            [10.0, 10.0],
            [-1.0, 1.0],
        ])
        array_2 = np.array([
            [1.0, 1.0],
            [-1.0, -1.0],
            [-1.0, 1.0],
        ])
        result = graph.connect_dense_prediction_graph(tf_user_representation=array_1,
                                                      tf_item_representation=array_2).eval(session=self.session)

        expected_result = np.array([
            [2.0, -2.0, 0.0],
            [20.0, -20.0, 0.0],
            [0.0, 0.0, 2.0]
        ])
        self.assertTrue(np.allclose(result, expected_result))

    def test_serial_prediction(self):
        graph = DotProductPredictionGraph()
        array_1 = np.array([
            [1.0, 1.0],
            [10.0, 10.0],
            [-1.0, 1.0],
        ])
        array_2 = np.array([
            [1.0, 1.0],
            [-1.0, -1.0],
            [-1.0, 1.0],
        ])

        x_user = np.array([0, 0, 0, 1, 1, 1, 2, 2, 2])
        x_item = np.array([0, 1, 2, 0, 1, 2, 0, 1, 2])

        result = graph.connect_serial_prediction_graph(tf_user_representation=array_1,
                                                       tf_item_representation=array_2,
                                                       tf_x_user=x_user,
                                                       tf_x_item=x_item,).eval(session=self.session)

        expected_result = np.array([2.0, -2.0, 0.0,
                                    20.0, -20.0, 0.0,
                                    0.0, 0.0, 2.0])
        self.assertTrue(np.allclose(result, expected_result))


class CosineSimilarityTestCase(TestCase):

    @classmethod
    def setUpClass(cls):
        cls.session = get_session()

    def test_dense_prediction(self):
        graph = CosineSimilarityPredictionGraph()
        array_1 = np.array([
            [1.0, 1.0],
            [10.0, 10.0],
            [-1.0, 1.0],
        ])
        array_2 = np.array([
            [1.0, 1.0],
            [-1.0, -1.0],
            [-1.0, 1.0],
        ])
        result = graph.connect_dense_prediction_graph(tf_user_representation=array_1,
                                                      tf_item_representation=array_2).eval(session=self.session)

        expected_result = np.array([
            [1.0, -1.0, 0.0],
            [1.0, -1.0, 0.0],
            [0.0, 0.0, 1.0]
        ])
        self.assertTrue(np.allclose(result, expected_result))

    def test_serial_prediction(self):
        graph = CosineSimilarityPredictionGraph()
        array_1 = np.array([
            [1.0, 1.0],
            [10.0, 10.0],
            [-1.0, 1.0],
        ])
        array_2 = np.array([
            [1.0, 1.0],
            [-1.0, -1.0],
            [-1.0, 1.0],
        ])

        x_user = np.array([0, 0, 0, 1, 1, 1, 2, 2, 2])
        x_item = np.array([0, 1, 2, 0, 1, 2, 0, 1, 2])

        result = graph.connect_serial_prediction_graph(tf_user_representation=array_1,
                                                       tf_item_representation=array_2,
                                                       tf_x_user=x_user,
                                                       tf_x_item=x_item,).eval(session=self.session)

        expected_result = np.array([1.0, -1.0, 0.0,
                                    1.0, -1.0, 0.0,
                                    0.0, 0.0, 1.0])
        self.assertTrue(np.allclose(result, expected_result))


class EuclideanSimilarityTestCase(TestCase):

    @classmethod
    def setUpClass(cls):
        cls.session = get_session()

    def test_dense_prediction(self):
        graph = EuclideanSimilarityPredictionGraph()
        array_1 = np.array([
            [1.0, 1.0],
            [2.0, 2.0],
            [-1.0, 1.0],
        ])
        array_2 = np.array([
            [1.0, 1.0],
            [-1.0, -1.0],
            [-1.0, 1.0],
        ])

        result = graph.connect_dense_prediction_graph(tf_user_representation=array_1,
                                                      tf_item_representation=array_2).eval(session=self.session)

        expected_result = np.array([
            [0.0, -math.sqrt(8.0), -2.0],
            [-math.sqrt(2.0), -math.sqrt(18.0), -math.sqrt(10.0)],
            [-2.0, -2.0, 0.0]]
        )
        self.assertTrue(np.allclose(result, expected_result))

    def test_serial_prediction(self):
        graph = EuclideanSimilarityPredictionGraph()
        array_1 = np.array([
            [1.0, 1.0],
            [2.0, 2.0],
            [-1.0, 1.0],
        ])
        array_2 = np.array([
            [1.0, 1.0],
            [-1.0, -1.0],
            [-1.0, 1.0],
        ])

        x_user = np.array([0, 0, 0, 1, 1, 1, 2, 2, 2])
        x_item = np.array([0, 1, 2, 0, 1, 2, 0, 1, 2])

        result = graph.connect_serial_prediction_graph(tf_user_representation=array_1,
                                                       tf_item_representation=array_2,
                                                       tf_x_user=x_user,
                                                       tf_x_item=x_item,).eval(session=self.session)

        expected_result = np.array([0.0, -math.sqrt(8.0), -2.0,
                                    -math.sqrt(2.0), -math.sqrt(18.0), -math.sqrt(10.0),
                                    -2.0, -2.0, 0.0])
        self.assertTrue(np.allclose(result, expected_result))



import numpy as np
import scipy.sparse as sp

from .kimrec import TensorRec


def precision_at_k(predicted_ranks, test_interactions, k=10, preserve_rows=False):
    """
    Modified from LightFM.
    :param predicted_ranks: Numpy matrix
    The results of model.predict_rank()
    :param test_interactions: scipy.sparse matrix
    Test interactions matrix of shape [n_users, n_items]
    :param k: int
    The rank at which to stop evaluating precision.
    :param preserve_rows: bool
    If True, a value of 0 will be returned for every user without test interactions.
    If False, only users with test interactions will be returned.
    :return: np.array
    """
    positive_test_interactions = test_interactions > 0
    ranks = sp.csr_matrix(predicted_ranks * positive_test_interactions.A)
    ranks.data = np.less(ranks.data, (k + 1), ranks.data)

    precision = np.squeeze(np.array(ranks.sum(axis=1))).astype(float) / k

    if not preserve_rows:
        precision = precision[positive_test_interactions.getnnz(axis=1) > 0]

    return precision


def recall_at_k(predicted_ranks, test_interactions, k=10, preserve_rows=False):
    """
    Modified from LightFM.
    :param predicted_ranks: Numpy matrix
    The results of model.predict_rank()
    :param test_interactions: scipy.sparse matrix
    Test interactions matrix of shape [n_users, n_items]
    :param k: int
    The rank at which to stop evaluating recall.
    :param preserve_rows: bool
    If True, a value of 0 will be returned for every user without test interactions.
    If False, only users with test interactions will be returned.
    :return: np.array
    """
    positive_test_interactions = test_interactions > 0
    ranks = sp.csr_matrix(predicted_ranks * positive_test_interactions.A)
    ranks.data = np.less(ranks.data, (k + 1), ranks.data)

    retrieved = np.squeeze(positive_test_interactions.getnnz(axis=1))
    hit = np.squeeze(np.array(ranks.sum(axis=1)))

    if not preserve_rows:
        hit = hit[positive_test_interactions.getnnz(axis=1) > 0]
        retrieved = retrieved[positive_test_interactions.getnnz(axis=1) > 0]

    return hit.astype(float) / retrieved.astype(float)


def _setup_ndcg(predicted_ranks, test_interactions, k=10):

    pos_inter = test_interactions > 0
    ror = sp.csr_matrix(predicted_ranks * pos_inter.A)

    relevance = sp.csr_matrix(test_interactions.A * pos_inter.A)

    k_mask = np.less(ror.data, k + 1)
    ror_at_k = np.maximum(np.multiply(ror.data, k_mask), 1)

    return relevance, k_mask, ror, ror_at_k


def _idcg(hits, k=10):
    sorted_hits = hits[np.argsort(-hits)][:min(len(hits), k)]
    idgc = np.sum((2**sorted_hits-1)/np.log2(np.arange(len(sorted_hits)) + 2))
    return idgc


def _dcg(relevance, k_mask, ror_at_k, ror):
    numer = (2**np.multiply(relevance.data, k_mask)) - 1
    denom = np.log2(ror_at_k + 1)
    ror.data = numer/denom  # ranks at 1
    dcg = ror.sum(axis=1).flatten()

    return dcg


def ndcg_at_k(predicted_ranks, test_interactions, k=10, preserve_rows=False):
    """
    Calculate Normalized Discounted Cumulative Gain @K.
    :param predicted_ranks: Numpy matrix
    The results of model.predict_rank()
    :param test_interactions: scipy.sparse matrix
    Test interactions matrix of shape [n_users, n_items]
    :param k: int
    The rank at which to stop evaluating NDCG.
    :param preserve_rows: bool
    If True, a value of 0 will be returned for every user without test interactions.
    If False, only users with test interactions will be returned.
    :return: np.array
    """

    relevance, k_mask, ranks_of_relevant, ror_at_k = _setup_ndcg(predicted_ranks,
                                                                 test_interactions,
                                                                 k)

    dcg = np.asarray(_dcg(relevance, k_mask, ror_at_k, ranks_of_relevant))[0]
    idcg = np.apply_along_axis(_idcg, 1, relevance.A)

    ndcg = dcg/idcg

    if not preserve_rows:
        positive_test_interactions = test_interactions > 0
        ndcg = ndcg[positive_test_interactions.getnnz(axis=1) > 0]

    return ndcg


def f1_score_at_k(predicted_ranks, test_interactions, k=10, preserve_rows=False):
    """
    :param model: TensorRec
    A trained TensorRec model.
    :param test_interactions: scipy.sparse matrix
    Test interactions matrix of shape [n_users, n_items]
    :param user_features: scipy.sparse matrix
    User features matrix of shape [n_users, n_user_features]
    :param item_features: scipy.sparse matrix
    Item features matrix of shape [n_items, n_item_features]
    :param k: int
    The rank at which to stop evaluating recall.
    :param preserve_rows: bool
    If True, a value of 0 will be returned for every user without test interactions.
    If False, only users with test interactions will be returned.
    :return: np.array
    """
    p_at_k = precision_at_k(predicted_ranks=predicted_ranks,
                            test_interactions=test_interactions,
                            k=k, preserve_rows=preserve_rows)
    r_at_k = recall_at_k(predicted_ranks=predicted_ranks,
                         test_interactions=test_interactions,
                         k=k, preserve_rows=preserve_rows)

    mean_p = np.mean(p_at_k)
    mean_r = np.mean(r_at_k)

    f1_score = (2.0 * mean_p * mean_r) / (mean_p + mean_r)
    return f1_score


def fit_and_eval(model, user_features, item_features, train_interactions, test_interactions, fit_kwargs, recall_k=30,
                 precision_k=5, ndcg_k=30):

    model.fit(user_features=user_features, item_features=item_features,
              interactions=train_interactions, **fit_kwargs)
    predicted_ranks = model.predict_rank(user_features=user_features, item_features=item_features)

    p_at_k = precision_at_k(predicted_ranks, test_interactions, k=precision_k)
    r_at_k = recall_at_k(predicted_ranks, test_interactions, k=recall_k)
    n_at_k = ndcg_at_k(predicted_ranks, test_interactions, k=ndcg_k)

    p_at_k_insample = precision_at_k(predicted_ranks, train_interactions, k=precision_k)
    r_at_k_insample = recall_at_k(predicted_ranks, train_interactions, k=recall_k)
    n_at_k_insample = ndcg_at_k(predicted_ranks, train_interactions, k=ndcg_k)

    return (np.mean(r_at_k), np.mean(p_at_k), np.mean(n_at_k), np.mean(r_at_k_insample), np.mean(p_at_k_insample),
            np.mean(n_at_k_insample))


def grid_check_model_on_dataset(train_interactions, test_interactions, user_features, item_features):

    results = []
    for n_components in [2, 4, 8, 16, 32, 64, 128, 256]:
        for epochs in [2, 4, 8, 16, 32, 64, 128, 256]:
            model = TensorRec(n_components=n_components)
            scores = fit_and_eval(model, user_features, item_features, train_interactions, test_interactions,
                                  fit_kwargs={'epochs': epochs})
            results.append((n_components, scores))
            print(n_components, epochs, scores)


def eval_random_ranks_on_dataset(interactions, recall_k=30, precision_k=5, ndcg_k=30):

    n_users, n_items = interactions.shape

    random_guesses = np.array([np.random.choice(a=n_items, size=n_items, replace=False) + 1 for _ in range(n_users)])

    p_at_k = precision_at_k(random_guesses, interactions, k=precision_k)
    r_at_k = recall_at_k(random_guesses, interactions, k=recall_k)
    n_at_k = ndcg_at_k(random_guesses, interactions, k=ndcg_k)

    return np.mean(r_at_k), np.mean(p_at_k), np.mean(n_at_k)



import math
import numpy as np
import random
import scipy.sparse as sp
import six
import tensorflow as tf
from weakref import WeakKeyDictionary

from .input_utils import create_tensorrec_dataset_from_sparse_matrix, create_tensorrec_dataset_from_tfrecord


def sample_items(n_items, n_users, n_sampled_items, replace):
    items_per_user = [np.random.choice(a=n_items, size=n_sampled_items, replace=replace)
                      for _ in range(n_users)]

    sample_indices = []
    for user, users_items in enumerate(items_per_user):
        for item in users_items:
            sample_indices.append((user, item))

    return np.array(sample_indices, np.int64)


def calculate_batched_alpha(num_batches, alpha):
    if num_batches < 1:
        raise ValueError('num_batches must be >=1, num_batches={}'.format(num_batches))
    elif num_batches > 1:
        batched_alpha = alpha / (math.e * math.log(num_batches))
    else:
        batched_alpha = alpha
    return batched_alpha


def datasets_from_raw_input(raw_input):

    if isinstance(raw_input, tf.data.Dataset):
        return [raw_input]

    if sp.issparse(raw_input):
        return [create_tensorrec_dataset_from_sparse_matrix(raw_input)]

    if isinstance(raw_input, six.string_types):
        return [create_tensorrec_dataset_from_tfrecord(raw_input)]

    if isinstance(raw_input, list):

        if all([isinstance(input_val, tf.data.Dataset) for input_val in raw_input]):
            return raw_input

        if all([sp.issparse(input_val) for input_val in raw_input]):
            return [create_tensorrec_dataset_from_sparse_matrix(input_sparse_matrix)
                    for input_sparse_matrix in raw_input]

        if all([isinstance(input_val, six.string_types) for input_val in raw_input]):
            return [create_tensorrec_dataset_from_tfrecord(input_str) for input_str in raw_input]

    raise ValueError('Input must be a scipy sparse matrix, an iterable of scipy sprase matrices, or a TensorFlow '
                     'Dataset')


def generate_dummy_data(num_users=15000, num_items=30000, interaction_density=.00045, num_user_features=200,
                        num_item_features=200, n_features_per_user=20, n_features_per_item=20,  pos_int_ratio=.5,
                        return_datasets=False):

    if pos_int_ratio <= 0.0:
        raise Exception("pos_int_ratio must be > 0")

    print("Generating positive interactions")
    interactions = sp.rand(num_users, num_items, density=interaction_density * pos_int_ratio)
    if pos_int_ratio < 1.0:
        print("Generating negative interactions")
        interactions += -1 * sp.rand(num_users, num_items, density=interaction_density * (1 - pos_int_ratio))

    print("Generating user features")
    user_features = sp.rand(num_users, num_user_features, density=float(n_features_per_user) / num_user_features)

    print("Generating item features")
    item_features = sp.rand(num_items, num_item_features, density=float(n_features_per_item) / num_item_features)

    if return_datasets:
        interactions = create_tensorrec_dataset_from_sparse_matrix(interactions)
        user_features = create_tensorrec_dataset_from_sparse_matrix(user_features)
        item_features = create_tensorrec_dataset_from_sparse_matrix(item_features)

    return interactions, user_features, item_features


def generate_dummy_data_with_indicator(num_users=15000, num_items=30000, interaction_density=.00045, pos_int_ratio=.5):

    n_user_features = int(num_users * 1.2)
    n_user_tags = num_users * 3
    n_item_features = int(num_items * 1.2)
    n_item_tags = num_items * 3
    n_interactions = (num_users * num_items) * interaction_density

    user_features = sp.lil_matrix((num_users, n_user_features))
    for i in range(num_users):
        user_features[i, i] = 1

    for i in range(n_user_tags):
        user_features[random.randrange(num_users), random.randrange(num_users, n_user_features)] = 1

    item_features = sp.lil_matrix((num_items, n_item_features))
    for i in range(num_items):
        item_features[i, i] = 1

    for i in range(n_item_tags):
        item_features[random.randrange(num_items), random.randrange(num_items, n_item_features)] = 1

    interactions = sp.lil_matrix((num_users, num_items))
    for i in range(int(n_interactions * pos_int_ratio)):
        interactions[random.randrange(num_users), random.randrange(num_items)] = 1

    for i in range(int(n_interactions * (1 - pos_int_ratio))):
        interactions[random.randrange(num_users), random.randrange(num_items)] = -1

    return interactions, user_features, item_features


def append_to_string_at_point(string, value, point):
    for _ in range(0, (point - len(string))):
        string += " "
    string += "{}".format(value)
    return string


def simple_tf_print(tensor, places=100):
    tf.compat.v1.disable_eager_execution()
    return tf.compat.v1.Print(tensor, [tensor, tf.shape(tensor)], summarize=places)


class lazyval(object):
    """
    Decorator that marks that an attribute of an instance should not be
    computed until needed, and that the value should be memoized.

    Lifted from quantopian/trading_calendars
    """
    def __init__(self, get):
        self._get = get
        self._cache = WeakKeyDictionary()

    def __get__(self, instance, owner):
        if instance is None:
            return self
        try:
            return self._cache[instance]
        except KeyError:
            self._cache[instance] = val = self._get(instance)
            return val

    def __set__(self, instance, value):
        raise AttributeError("Can't set read-only attribute.")

    def __delitem__(self, instance):
        del self._cache[instance]



from unittest import TestCase
from numpy.random import shuffle
import numpy as np
import scipy.sparse as sps

from tensorrec import TensorRec
from tensorrec.eval import recall_at_k, precision_at_k, f1_score_at_k, ndcg_at_k
from tensorrec.eval import _setup_ndcg, _idcg, _dcg
from tensorrec.util import generate_dummy_data_with_indicator


class EvalTestCase(TestCase):

    @classmethod
    def setUpClass(cls):
        cls.interactions, cls.user_features, cls.item_features = generate_dummy_data_with_indicator(
            num_users=10, num_items=12, interaction_density=.5)
        model = TensorRec(n_components=10)
        model.fit(cls.interactions, cls.user_features, cls.item_features, epochs=10)
        cls.model = model
        cls.ranks = model.predict_rank(user_features=cls.user_features, item_features=cls.item_features)

    def test_recall_at_k(self):
        # Check non-none results with and without preserve_rows
        self.assertIsNotNone(recall_at_k(predicted_ranks=self.ranks,
                                         test_interactions=self.interactions,
                                         k=5,
                                         preserve_rows=False))
        self.assertIsNotNone(recall_at_k(predicted_ranks=self.ranks,
                                         test_interactions=self.interactions,
                                         k=5,
                                         preserve_rows=True))

    def test_precision_at_k(self):
        # Check non-none results with and without preserve_rows
        self.assertIsNotNone(precision_at_k(predicted_ranks=self.ranks,
                                            test_interactions=self.interactions,
                                            k=5,
                                            preserve_rows=False))
        self.assertIsNotNone(precision_at_k(predicted_ranks=self.ranks,
                                            test_interactions=self.interactions,
                                            k=5,
                                            preserve_rows=True))

    def test_f1_score_at_k(self):
        # Check non-none results with and without preserve_rows
        self.assertIsNotNone(f1_score_at_k(predicted_ranks=self.ranks,
                                           test_interactions=self.interactions,
                                           k=5,
                                           preserve_rows=False))
        self.assertIsNotNone(f1_score_at_k(predicted_ranks=self.ranks,
                                           test_interactions=self.interactions,
                                           k=5,
                                           preserve_rows=True))

    def test_ndcg_at_k(self):

        ndcg10 = np.mean(ndcg_at_k(
            predicted_ranks=self.ranks,
            test_interactions=self.interactions,
            k=10,
            preserve_rows=False
        ))

        ndcg20 = np.mean(ndcg_at_k(
            predicted_ranks=self.ranks,
            test_interactions=self.interactions,
            k=20,
            preserve_rows=False
        ))

        ndcg5 = np.mean(ndcg_at_k(
            predicted_ranks=self.ranks,
            test_interactions=self.interactions,
            k=5,
            preserve_rows=False
        ))

        self.assertIsNotNone(ndcg10)
        self.assertGreaterEqual(ndcg10, ndcg5)
        self.assertEqual(ndcg20, ndcg10)
        self.assertLess(ndcg20, 1)

    def test_idcg_at_k(self):

        ordered = np.array([3, 3, 3, 2, 2, 2, 1, 0])
        wiki_idcg = np.array([3, 3, 3, 2, 2, 2, 1, 0])

        # note different than on page calculation, they use linear
        shuffle(wiki_idcg)
        exponential_calc = _idcg(wiki_idcg)
        self.assertAlmostEqual(exponential_calc, 18.77105, 3)

        ordered_calc = _idcg(ordered)
        self.assertEqual(ordered_calc, exponential_calc)

        binary = np.array([1, 1, 1, 1, 0, 0])
        # By hand
        terms = [(2**e-1)/(np.log2(i+2)) for i, e in enumerate(list(binary))]
        est_idcg = np.sum(terms)
        hits = binary
        self.assertEqual(_idcg(hits), est_idcg)

    def test_ndcg_setup(self):
        wiki_rel = sps.lil_matrix(np.array([3, 2, 3, 0, 1, 2]))
        wiki_rank = np.array([1, 2, 3, 4, 5, 6])

        # at 10
        rel, k_mask, ror, ror_at_k = _setup_ndcg(wiki_rank, wiki_rel)

        self.assertIsNotNone(rel)
        self.assertEqual(len(k_mask), 5)  # all non-zero elements in relevance where rank <= k
        self.assertEqual(len(k_mask), len(ror_at_k))
        ror_cnt = sum([v[0] == v[1] for v in zip(ror.data, [1, 2, 3, 5, 6])])
        self.assertEqual(ror_cnt, len(k_mask))  # check length is k mask and correct index

    def test_dcg(self):
        wiki_rel = sps.lil_matrix(np.array([3, 3, 1, 0, 2]))
        wiki_rank = np.array([1, 2, 3, 4, 5])
        rel, k_mask, ror, ror_at_k = _setup_ndcg(wiki_rank, wiki_rel)

        # by hand...
        mat = np.array([3, 3, 1, 2])
        ranks = np.array([1, 2, 3, 5])

        numer_bh = 2**np.multiply(mat, [1, 1, 1, 1]) - 1
        denom_bh = np.log2(ranks + 1)

        # in function calc
        numer_func = (2 ** np.multiply(rel.data, k_mask)) - 1
        denom_func = np.log2(ror_at_k + 1)

        numer_all_eq = sum([v[0] == v[1] for v in zip(numer_bh, numer_func)])
        self.assertEqual(numer_all_eq, len(mat))

        denom_all_eq = sum([v[0] == v[1] for v in zip(denom_bh, denom_func)])
        self.assertEqual(denom_all_eq, len(mat))

        by_hand_dcg = np.sum(numer_bh/denom_bh)
        func_dcg = _dcg(rel, k_mask, ror_at_k, ror)
        self.assertEqual(by_hand_dcg, func_dcg)

        # functional NDCG should equal example another example
        # http://people.cs.georgetown.edu/~nazli/classes/ir-Slides/Evaluation-13.pdf
        # about .96 but since we use Burges 2005 formulation
        # it will be a little different
        v = func_dcg/_idcg(np.array([3, 3, 1, 0, 2]))
        self.assertAlmostEqual(v.item(0), .979762, 3)



from .util import lazyval


class TensorRecException(Exception):
    msg = None

    def __init__(self, **kwargs):
        self.kwargs = kwargs

    @lazyval
    def message(self):
        return str(self)

    def __str__(self):
        msg = self.msg.format(**self.kwargs)
        return msg

    __unicode__ = __str__
    __repr__ = __str__


class ModelNotBiasedException(TensorRecException):
    msg = 'Cannot predict {actor} bias for unbiased model'


class ModelNotFitException(TensorRecException):
    msg = "{method}() has been called before model fitting. Call fit() or fit_partial() before calling {method}()."


class ModelWithoutAttentionException(TensorRecException):
    msg = "This TensorRec model does not use attention. Try re-building TensorRec with a valid 'attention_graph' arg."


class BatchNonSparseInputException(TensorRecException):
    msg = 'In order to support user batching at fit time, interactions and user_features must both be scipy.sparse ' \
          'matrices.'


class TfVersionException(TensorRecException):
    msg = "You need to have at least TensorFlow version 1.7 installed in order to use TensorRec properly. You have " \
          "currently installed TensorFlow: {tf_version}"



from .kimrec import TensorRec
from . import eval
from . import input_utils
from . import loss_graphs
from . import representation_graphs
from . import prediction_graphs
from . import session_management
from . import util

__version__ = '0.3.0'

__all__ = [
    TensorRec, eval, util, loss_graphs, representation_graphs, prediction_graphs, session_management, input_utils
]

# Suppress TensorFlow logs
import logging
logging.getLogger('tensorflow').disabled = True



import abc
import tensorflow as tf


class AbstractLossGraph(object):
    __metaclass__ = abc.ABCMeta

    # If True, dense prediction results will be passed to the loss function
    is_dense = False

    # If True, randomly sampled predictions will be passed to the loss function
    is_sample_based = False
    # If True, and if is_sample_based is True, predictions will be sampled with replacement
    is_sampled_with_replacement = False

    @abc.abstractmethod
    def connect_loss_graph(self, tf_prediction_serial, tf_interactions_serial, tf_interactions, tf_n_users, tf_n_items,
                           tf_prediction, tf_rankings, tf_sample_predictions, tf_n_sampled_items):
        """
        This method is responsible for consuming a number of possible nodes from the graph and calculating loss from
        those nodes.

        The following parameters are always passed in
        :param tf_prediction_serial: tf.Tensor
        The recommendation scores as a Tensor of shape [n_samples, 1]
        :param tf_interactions_serial: tf.Tensor
        The sample interactions corresponding to tf_prediction_serial as a Tensor of shape [n_samples, 1]
        :param tf_interactions: tf.SparseTensor
        The sample interactions as a SparseTensor of shape [n_users, n_items]
        :param tf_n_users: tf.placeholder
        The number of users in tf_interactions
        :param tf_n_items: tf.placeholder
        The number of items in tf_interactions

        The following parameters are passed in if is_dense is True
        :param tf_prediction: tf.Tensor
        The recommendation scores as a Tensor of shape [n_users, n_items]
        :param tf_rankings: tf.Tensor
        The item ranks as a Tensor of shape [n_users, n_items]

        The following parameters are passed in if is_sample_based is True
        :param tf_sample_predictions: tf.Tensor
        The recommendation scores of a sample of items of shape [n_users, n_sampled_items]
        :param tf_n_sampled_items: tf.placeholder
        The number of items per user in tf_sample_predictions

        :return: tf.Tensor
        The loss value.
        """
        pass


class RMSELossGraph(AbstractLossGraph):
    """
    This loss function returns the root mean square error between the predictions and the true interactions.
    Interactions can be any positive or negative values, and this loss function is sensitive to magnitude.
    """
    def connect_loss_graph(self, tf_prediction_serial, tf_interactions_serial, **kwargs):
        return tf.sqrt(tf.reduce_mean(tf.square(tf_interactions_serial - tf_prediction_serial)))


class RMSEDenseLossGraph(AbstractLossGraph):
    """
    This loss function returns the root mean square error between the predictions and the true interactions, including
    all non-interacted values as 0s.
    Interactions can be any positive or negative values, and this loss function is sensitive to magnitude.
    """
    is_dense = True

    def connect_loss_graph(self, tf_interactions, tf_prediction, **kwargs):
        tf.compat.v1.disable_eager_execution()
        error = tf.compat.v1.sparse_add(tf_interactions, -1.0 * tf_prediction)
        return tf.sqrt(tf.reduce_mean(tf.square(error)))


class SeparationLossGraph(AbstractLossGraph):
    """
    This loss function models the explicit positive and negative interaction predictions as normal distributions and
    returns the probability of overlap between the two distributions.
    Interactions can be any positive or negative values, but this loss function ignores the magnitude of the
    interaction -- interactions are grouped in to {i <= 0} and {i > 0}.
    """
    def connect_loss_graph(self, tf_prediction_serial, tf_interactions_serial, **kwargs):

        tf_positive_mask = tf.greater(tf_interactions_serial, 0.0)
        tf_negative_mask = tf.less_equal(tf_interactions_serial, 0.0)

        tf_positive_predictions = tf.boolean_mask(tf_prediction_serial, tf_positive_mask)
        tf_negative_predictions = tf.boolean_mask(tf_prediction_serial, tf_negative_mask)

        tf_pos_mean, tf_pos_var = tf.nn.moments(tf_positive_predictions, axes=[0])
        tf_neg_mean, tf_neg_var = tf.nn.moments(tf_negative_predictions, axes=[0])
        # tensorflow version error solution : v1->v2 ("compat.v1" doesn't work)
        %tensorflow_version 1.x
        tf_overlap_distribution = tf.contrib.distributions.Normal(loc=(tf_neg_mean - tf_pos_mean),
                                                                  scale=tf.sqrt(tf_neg_var + tf_pos_var))

        loss = 1.0 - tf_overlap_distribution.cdf(0.0)
        return loss


class SeparationDenseLossGraph(AbstractLossGraph):
    """
    This loss function models all positive and negative interaction predictions as normal distributions and
    returns the probability of overlap between the two distributions. This loss function includes non-interacted items
    as negative interactions.
    Interactions can be any positive or negative values, but this loss function ignores the magnitude of the
    interaction -- interactions are grouped in to {i <= 0} and {i > 0}.
    """
    is_dense = True

    def connect_loss_graph(self, tf_prediction, tf_interactions, **kwargs):

        interactions_shape = tf.shape(tf_interactions)
        int_serial_shape = tf.cast([interactions_shape[0] * interactions_shape[1]], tf.int32)
        tf_interactions_serial = tf.reshape(tf.sparse_tensor_to_dense(tf_interactions),
                                            shape=int_serial_shape)

        prediction_shape = tf.shape(tf_prediction)
        pred_serial_shape = tf.cast([prediction_shape[0] * prediction_shape[1]], tf.int32)
        tf_prediction_serial = tf.reshape(tf_prediction, shape=pred_serial_shape)

        tf_positive_mask = tf.greater(tf_interactions_serial, 0.0)
        tf_negative_mask = tf.less_equal(tf_interactions_serial, 0.0)

        tf_positive_predictions = tf.boolean_mask(tf_prediction_serial, tf_positive_mask)
        tf_negative_predictions = tf.boolean_mask(tf_prediction_serial, tf_negative_mask)

        tf_pos_mean, tf_pos_var = tf.nn.moments(tf_positive_predictions, axes=[0])
        tf_neg_mean, tf_neg_var = tf.nn.moments(tf_negative_predictions, axes=[0])
        # tensorflow version error solution : v1->v2 ("compat.v1" doesn't work)
        %tensorflow_version 1.x
        tf_overlap_distribution = tf.contrib.distributions.Normal(loc=(tf_neg_mean - tf_pos_mean),
                                                                  scale=tf.sqrt(tf_neg_var + tf_pos_var))

        loss = 1.0 - tf_overlap_distribution.cdf(0.0)
        return loss


class WMRBLossGraph(AbstractLossGraph):
    """
    Approximation of http://ceur-ws.org/Vol-1905/recsys2017_poster3.pdf
    Interactions can be any positive values, but magnitude is ignored. Negative interactions are ignored.
    """
    is_sample_based = True

    def connect_loss_graph(self, tf_prediction_serial, tf_interactions, tf_sample_predictions, tf_n_items,
                           tf_n_sampled_items, **kwargs):

        return self.weighted_margin_rank_batch(tf_prediction_serial=tf_prediction_serial,
                                               tf_interactions=tf_interactions,
                                               tf_sample_predictions=tf_sample_predictions,
                                               tf_n_items=tf_n_items,
                                               tf_n_sampled_items=tf_n_sampled_items)

    def weighted_margin_rank_batch(self, tf_prediction_serial, tf_interactions, tf_sample_predictions, tf_n_items,
                                   tf_n_sampled_items):
        positive_interaction_mask = tf.greater(tf_interactions.values, 0.0)
        positive_interaction_indices = tf.boolean_mask(tf_interactions.indices,
                                                       positive_interaction_mask)

        # [ n_positive_interactions ]
        positive_predictions = tf.boolean_mask(tf_prediction_serial,
                                               positive_interaction_mask)

        n_items = tf.cast(tf_n_items, dtype=tf.float32)
        n_sampled_items = tf.cast(tf_n_sampled_items, dtype=tf.float32)

        # [ n_positive_interactions, n_sampled_items ]
        mapped_predictions_sample_per_interaction = tf.gather(params=tf_sample_predictions,
                                                              indices=tf.transpose(positive_interaction_indices)[0])

        # [ n_positive_interactions, n_sampled_items ]
        summation_term = tf.maximum(1.0
                                    - tf.expand_dims(positive_predictions, axis=1)
                                    + mapped_predictions_sample_per_interaction,
                                    0.0)

        # [ n_positive_interactions ]
        sampled_margin_rank = (n_items / n_sampled_items) * tf.reduce_sum(summation_term, axis=1)
        
        tf.compat.v1.disable_eager_execution()
        loss = tf.compat.v1.log(sampled_margin_rank + 1.0)
        return loss


class BalancedWMRBLossGraph(WMRBLossGraph):
    """
    This loss graph extends WMRB by making it sensitive to interaction magnitude and weighting the loss of each item by
    1 / sum(interactions) per item.
    Interactions can be any positive values. Negative interactions are ignored.
    """
    def weighted_margin_rank_batch(self, tf_prediction_serial, tf_interactions, tf_sample_predictions, tf_n_items,
                                   tf_n_sampled_items):
        positive_interaction_mask = tf.greater(tf_interactions.values, 0.0)
        positive_interaction_indices = tf.boolean_mask(tf_interactions.indices,
                                                       positive_interaction_mask)
        positive_interaction_values = tf.boolean_mask(tf_interactions.values,
                                                      positive_interaction_mask)

        positive_interactions = tf.SparseTensor(indices=positive_interaction_indices,
                                                values=positive_interaction_values,
                                                dense_shape=tf_interactions.dense_shape)
                                                
        tf.compat.v1.disable_eager_execution()
        listening_sum_per_item = tf.compat.v1.sparse_reduce_sum(positive_interactions, axis=0)
        gathered_sums = tf.gather(params=listening_sum_per_item,
                                  indices=tf.transpose(positive_interaction_indices)[1])

        # [ n_positive_interactions ]
        positive_predictions = tf.boolean_mask(tf_prediction_serial,
                                               positive_interaction_mask)

        n_items = tf.cast(tf_n_items, dtype=tf.float32)
        n_sampled_items = tf.cast(tf_n_sampled_items, dtype=tf.float32)

        # [ n_positive_interactions, n_sampled_items ]
        mapped_predictions_sample_per_interaction = tf.gather(params=tf_sample_predictions,
                                                              indices=tf.transpose(positive_interaction_indices)[0])

        # [ n_positive_interactions, n_sampled_items ]
        summation_term = tf.maximum(1.0
                                    - tf.expand_dims(positive_predictions, axis=1)
                                    + mapped_predictions_sample_per_interaction,
                                    0.0)

        # [ n_positive_interactions ]
        sampled_margin_rank = ((n_items / n_sampled_items)
                               * tf.reduce_sum(summation_term, axis=1)
                               * positive_interaction_values / gathered_sums)
                               
        tf.compat.v1.disable_eager_execution()
        loss = tf.compat.v1.log(sampled_margin_rank + 1.0)
        return loss



from unittest import TestCase


class ReadmeTestCase(TestCase):

    def test_basic_usage(self):
        import numpy as np
        import tensorrec

        # Build the model with default parameters
        model = tensorrec.TensorRec()

        # Generate some dummy data
        interactions, user_features, item_features = tensorrec.util.generate_dummy_data(num_users=100,
                                                                                        num_items=150,
                                                                                        interaction_density=.05)

        # Fit the model for 5 epochs
        model.fit(interactions, user_features, item_features, epochs=5, verbose=True)

        # Predict scores and ranks for all users and all items
        predictions = model.predict(user_features=user_features,
                                    item_features=item_features)
        predicted_ranks = model.predict_rank(user_features=user_features,
                                             item_features=item_features)

        # Calculate and print the recall at 10
        r_at_k = tensorrec.eval.recall_at_k(predicted_ranks, interactions, k=10)
        print(np.mean(r_at_k))

        self.assertIsNotNone(predictions)

    def test_custom_repr_graph(self):
        import tensorflow as tf
        import tensorrec

        # Define a custom representation function graph
        class TanhRepresentationGraph(tensorrec.representation_graphs.AbstractRepresentationGraph):
            def connect_representation_graph(self, tf_features, n_components, n_features, node_name_ending):
                """
                This representation function embeds the user/item features by passing them through a single tanh layer.
                :param tf_features: tf.SparseTensor
                The user/item features as a SparseTensor of dimensions [n_users/items, n_features]
                :param n_components: int
                The dimensionality of the resulting representation.
                :param n_features: int
                The number of features in tf_features
                :param node_name_ending: String
                Either 'user' or 'item'
                :return:
                A tuple of (tf.Tensor, list) where the first value is the resulting representation in n_components
                dimensions and the second value is a list containing all tf.Variables which should be subject to
                regularization.
                """
                tf_tanh_weights = tf.Variable(tf.random_normal([n_features, n_components], stddev=.5),
                                              name='tanh_weights_%s' % node_name_ending)

                tf_repr = tf.nn.tanh(tf.sparse_tensor_dense_matmul(tf_features, tf_tanh_weights))

                # Return repr layer and variables
                return tf_repr, [tf_tanh_weights]

        # Build a model with the custom representation function
        model = tensorrec.TensorRec(user_repr_graph=TanhRepresentationGraph(),
                                    item_repr_graph=TanhRepresentationGraph())

        # Generate some dummy data
        interactions, user_features, item_features = tensorrec.util.generate_dummy_data(num_users=100,
                                                                                        num_items=150,
                                                                                        interaction_density=.05)

        # Fit the model for 5 epochs
        model.fit(interactions, user_features, item_features, epochs=5, verbose=True)

        self.assertIsNotNone(model)

    def test_custom_loss_graph(self):
        import tensorflow as tf
        import tensorrec

        # Define a custom loss graph
        class SimpleLossGraph(tensorrec.loss_graphs.AbstractLossGraph):
            def connect_loss_graph(self, tf_prediction_serial, tf_interactions_serial, **kwargs):
                """
                This loss function returns the absolute simple error between the predictions and the interactions.
                :param tf_prediction_serial: tf.Tensor
                The recommendation scores as a Tensor of shape [n_samples, 1]
                :param tf_interactions_serial: tf.Tensor
                The sample interactions corresponding to tf_prediction_serial as a Tensor of shape [n_samples, 1]
                :param kwargs:
                Other TensorFlow nodes.
                :return:
                A tf.Tensor containing the learning loss.
                """
                return tf.reduce_mean(tf.abs(tf_interactions_serial - tf_prediction_serial))

        # Build a model with the custom loss function
        model = tensorrec.TensorRec(loss_graph=SimpleLossGraph())

        # Generate some dummy data
        interactions, user_features, item_features = tensorrec.util.generate_dummy_data(num_users=100,
                                                                                        num_items=150,
                                                                                        interaction_density=.05)

        # Fit the model for 5 epochs
        model.fit(interactions, user_features, item_features, epochs=5, verbose=True)

        self.assertIsNotNone(model)



from unittest import TestCase

from tensorrec.util import calculate_batched_alpha


class UtilTestcase(TestCase):

    def test_calculate_batched_alpha(self):

        # Test no-transformation
        n_batches = 1
        alpha = .01
        self.assertEqual(alpha, calculate_batched_alpha(num_batches=n_batches, alpha=alpha))

        # Test two batches
        n_batches = 2
        alpha = .01
        self.assertAlmostEqual(.53074 * alpha,
                               calculate_batched_alpha(num_batches=n_batches, alpha=alpha),
                               places=5)

        # Test a bad number of batches
        with self.assertRaises(ValueError):
            calculate_batched_alpha(num_batches=0, alpha=alpha)



import numpy as np
import os
import shutil
import tempfile
from unittest import TestCase

import tensorflow as tf

from tensorrec import TensorRec
from tensorrec.errors import (
    ModelNotBiasedException, ModelNotFitException, ModelWithoutAttentionException, BatchNonSparseInputException
)
from tensorrec.input_utils import create_tensorrec_dataset_from_sparse_matrix, write_tfrecord_from_sparse_matrix
from tensorrec.representation_graphs import NormalizedLinearRepresentationGraph, LinearRepresentationGraph
from tensorrec.session_management import set_session
from tensorrec.util import generate_dummy_data


class TensorRecTestCase(TestCase):

    @classmethod
    def setUpClass(cls):

        cls.n_user_features = 200
        cls.n_item_features = 150

        cls.interactions, cls.user_features, cls.item_features = generate_dummy_data(
            num_users=15, num_items=30, interaction_density=.5, num_user_features=cls.n_user_features,
            num_item_features=cls.n_item_features, n_features_per_user=20, n_features_per_item=20, pos_int_ratio=.5
        )

        set_session(None)
        cls.temp_dir = tempfile.mkdtemp()
        cls.interactions_path = os.path.join(cls.temp_dir, 'interactions.tfrecord')
        cls.user_features_path = os.path.join(cls.temp_dir, 'user_features.tfrecord')
        cls.item_features_path = os.path.join(cls.temp_dir, 'item_features.tfrecord')

        write_tfrecord_from_sparse_matrix(cls.user_features_path, cls.user_features)
        write_tfrecord_from_sparse_matrix(cls.item_features_path, cls.item_features)
        write_tfrecord_from_sparse_matrix(cls.interactions_path, cls.interactions)

    @classmethod
    def tearDownClass(cls):
        shutil.rmtree(cls.temp_dir)

    def test_init(self):
        self.assertIsNotNone(TensorRec())

    def test_init_fail_0_components(self):
        with self.assertRaises(ValueError):
            TensorRec(n_components=0)

    def test_init_fail_none_factory(self):
        with self.assertRaises(ValueError):
            TensorRec(user_repr_graph=None)
        with self.assertRaises(ValueError):
            TensorRec(item_repr_graph=None)
        with self.assertRaises(ValueError):
            TensorRec(loss_graph=None)

    def test_init_fail_bad_loss_graph(self):
        with self.assertRaises(ValueError):
            TensorRec(loss_graph=np.mean)

    def test_init_fail_attention_with_1_taste(self):
        with self.assertRaises(ValueError):
            TensorRec(n_tastes=1, attention_graph=LinearRepresentationGraph())

    def test_init_fail_bad_attention_graph(self):
        with self.assertRaises(ValueError):
            TensorRec(attention_graph=np.mean)

    def test_predict_fail_unfit(self):
        model = TensorRec()
        with self.assertRaises(ModelNotFitException):
            model.predict(self.user_features, self.item_features)
        with self.assertRaises(ModelNotFitException):
            model.predict_rank(self.user_features, self.item_features)

        with self.assertRaises(ModelNotFitException):
            model.predict_user_representation(self.user_features)
        with self.assertRaises(ModelNotFitException):
            model.predict_item_representation(self.item_features)
        with self.assertRaises(ModelNotFitException):
            model.predict_user_attention_representation(self.user_features)

        with self.assertRaises(ModelNotFitException):
            model.predict_similar_items(self.item_features, item_ids=[1], n_similar=5)

        with self.assertRaises(ModelNotFitException):
            model.predict_item_bias(self.item_features)
        with self.assertRaises(ModelNotFitException):
            model.predict_user_bias(self.user_features)

    def test_fit_verbose(self):
        model = TensorRec(n_components=10)
        model.fit(self.interactions, self.user_features, self.item_features, epochs=10, verbose=True)
        # Ensure that the nodes have been built
        self.assertIsNotNone(model.tf_prediction)

    def test_fit_batched(self):
        model = TensorRec(n_components=10)
        model.fit(self.interactions, self.user_features, self.item_features, epochs=10, user_batch_size=2)
        # Ensure that the nodes have been built
        self.assertIsNotNone(model.tf_prediction)

    def test_fit_fail_bad_input(self):
        model = TensorRec(n_components=10)
        with self.assertRaises(ValueError):
            model.fit(np.array([1, 2, 3, 4]), self.user_features, self.item_features, epochs=10)
        with self.assertRaises(ValueError):
            model.fit(self.interactions, np.array([1, 2, 3, 4]), self.item_features, epochs=10)
        with self.assertRaises(ValueError):
            model.fit(self.interactions, self.user_features, np.array([1, 2, 3, 4]), epochs=10)

    def test_fit_fail_mismatched_batches(self):
        model = TensorRec(n_components=10)
        with self.assertRaises(ValueError):
            model.fit(self.interactions,
                      [self.user_features, self.user_features],
                      [self.item_features, self.item_features, self.item_features],
                      epochs=10)

        with self.assertRaises(ValueError):
            model.fit(self.interactions,
                      [self.user_features, self.user_features],
                      [self.item_features, self.item_features],
                      epochs=10)

        model.fit([self.interactions, self.interactions],
                  [self.user_features, self.user_features],
                  self.item_features,
                  epochs=10)

        model.fit([self.interactions, self.interactions],
                  [self.user_features, self.user_features],
                  [self.item_features, self.item_features],
                  epochs=10)

    def test_fit_fail_batching_dataset(self):
        model = TensorRec(n_components=10)

        interactions_as_dataset = create_tensorrec_dataset_from_sparse_matrix(self.interactions)
        with self.assertRaises(BatchNonSparseInputException):
            model.fit(interactions_as_dataset, self.user_features, self.item_features, epochs=10, user_batch_size=2)

    def test_fit_user_feature_as_dataset(self):
        uf_as_dataset = create_tensorrec_dataset_from_sparse_matrix(self.user_features)
        model = TensorRec(n_components=10)
        model.fit(self.interactions, uf_as_dataset, self.item_features, epochs=10)

    def test_fit_item_feature_as_dataset(self):
        if_as_dataset = create_tensorrec_dataset_from_sparse_matrix(self.item_features)
        model = TensorRec(n_components=10)
        model.fit(self.interactions, self.user_features, if_as_dataset, epochs=10)

    def test_fit_interactions_as_dataset(self):
        int_as_dataset = create_tensorrec_dataset_from_sparse_matrix(self.interactions)
        model = TensorRec(n_components=10)
        model.fit(int_as_dataset, self.user_features, self.item_features, epochs=10)

    def test_fit_from_datasets(self):
        uf_as_dataset = create_tensorrec_dataset_from_sparse_matrix(self.user_features)
        if_as_dataset = create_tensorrec_dataset_from_sparse_matrix(self.item_features)
        int_as_dataset = create_tensorrec_dataset_from_sparse_matrix(self.interactions)
        model = TensorRec(n_components=10)
        model.fit(int_as_dataset, uf_as_dataset, if_as_dataset, epochs=10)

    def test_fit_from_tfrecords(self):
        set_session(None)
        model = TensorRec(n_components=10)
        model.fit(self.interactions_path, self.user_features_path, self.item_features_path, epochs=10)


class TensorRecAPITestCase(TestCase):

    @classmethod
    def setUpClass(cls):

        cls.n_users = 15
        cls.n_items = 30

        cls.interactions, cls.user_features, cls.item_features = generate_dummy_data(
            num_users=cls.n_users, num_items=cls.n_items, interaction_density=.5, num_user_features=200,
            num_item_features=200, n_features_per_user=20, n_features_per_item=20, pos_int_ratio=.5
        )

        cls.standard_model = TensorRec(n_components=10)
        cls.standard_model.fit(cls.interactions, cls.user_features, cls.item_features, epochs=10)

        cls.unbiased_model = TensorRec(n_components=10, biased=False)
        cls.unbiased_model.fit(cls.interactions, cls.user_features, cls.item_features, epochs=10)

    def test_fit(self):
        # Ensure that the nodes have been built
        self.assertIsNotNone(self.standard_model.tf_prediction)

    def test_predict(self):
        predictions = self.standard_model.predict(user_features=self.user_features,
                                                  item_features=self.item_features)

        self.assertEqual(predictions.shape, (self.n_users, self.n_items))

    def test_predict_rank(self):
        ranks = self.standard_model.predict_rank(user_features=self.user_features,
                                                 item_features=self.item_features)

        self.assertEqual(ranks.shape, (self.n_users, self.n_items))
        for x in range(ranks.shape[0]):
            for y in range(ranks.shape[1]):
                val = ranks[x][y]
                self.assertGreater(val, 0)

    def test_predict_similar_items(self):
        sims = self.standard_model.predict_similar_items(item_features=self.item_features,
                                                         item_ids=[6, 12],
                                                         n_similar=5)

        # Two items, two rows of sims
        self.assertEqual(len(sims), 2)

        for item_sims in sims:
            # Should equal n_similar
            self.assertEqual(len(item_sims), 5)

    def test_fit_predict_unbiased(self):
        predictions = self.unbiased_model.predict(user_features=self.user_features, item_features=self.item_features)
        self.assertEqual(predictions.shape, (self.n_users, self.n_items))

    def test_predict_user_repr(self):
        user_repr = self.unbiased_model.predict_user_representation(self.user_features)
        self.assertEqual(user_repr.shape, (self.n_users, 10))

    def test_predict_item_repr(self):
        item_repr = self.unbiased_model.predict_item_representation(self.item_features)
        self.assertEqual(item_repr.shape, (self.n_items, 10))

    def test_predict_user_bias_unbiased_model(self):
        self.assertRaises(
            ModelNotBiasedException,
            self.unbiased_model.predict_user_bias,
            self.user_features)

    def test_predict_item_bias_unbiased_model(self):
        self.assertRaises(
            ModelNotBiasedException,
            self.unbiased_model.predict_item_bias,
            self.item_features)

    def test_predict_user_attn_repr(self):
        # This test will be overwritten by the tests that have attention
        with self.assertRaises(ModelWithoutAttentionException):
            self.unbiased_model.predict_user_attention_representation(self.user_features)


class TensorRecBiasedPrediction(TestCase):
    # TODO: Collapse these into TensorRecTestCase once the fit bug is fixed

    @classmethod
    def setUpClass(cls):
        cls.interactions, cls.user_features, cls.item_features = generate_dummy_data(
            num_users=15, num_items=30, interaction_density=.5, num_user_features=200, num_item_features=200,
            n_features_per_user=20, n_features_per_item=20, pos_int_ratio=.5
        )

        cls.standard_model = TensorRec(n_components=10)
        cls.standard_model.fit(cls.interactions, cls.user_features, cls.item_features, epochs=10)

    def test_predict_user_bias(self):
        user_bias = self.standard_model.predict_user_bias(self.user_features)
        self.assertTrue(any(user_bias))  # Make sure it isn't all 0s

    def test_predict_item_bias(self):
        item_bias = self.standard_model.predict_item_bias(self.item_features)
        self.assertTrue(any(item_bias))  # Make sure it isn't all 0s


class TensorRecAPINTastesTestCase(TensorRecAPITestCase):

    @classmethod
    def setUpClass(cls):

        cls.n_users = 15
        cls.n_items = 30

        cls.interactions, cls.user_features, cls.item_features = generate_dummy_data(
            num_users=cls.n_users, num_items=cls.n_items, interaction_density=.5, num_user_features=200,
            num_item_features=200, n_features_per_user=20, n_features_per_item=20, pos_int_ratio=.5
        )

        cls.standard_model = TensorRec(n_components=10,
                                       n_tastes=3,
                                       user_repr_graph=NormalizedLinearRepresentationGraph())
        cls.standard_model.fit(cls.interactions, cls.user_features, cls.item_features, epochs=10)

        cls.unbiased_model = TensorRec(n_components=10,
                                       n_tastes=3,
                                       biased=False,
                                       user_repr_graph=NormalizedLinearRepresentationGraph())
        cls.unbiased_model.fit(cls.interactions, cls.user_features, cls.item_features, epochs=10)

    def test_predict_user_repr(self):
        user_repr = self.unbiased_model.predict_user_representation(self.user_features)

        # 3 tastes, shape[0] users, 10 components
        self.assertEqual(user_repr.shape, (3, self.user_features.shape[0], 10))


class TensorRecAPIAttentionTestCase(TensorRecAPINTastesTestCase):

    @classmethod
    def setUpClass(cls):

        cls.n_users = 15
        cls.n_items = 30

        cls.interactions, cls.user_features, cls.item_features = generate_dummy_data(
            num_users=cls.n_users, num_items=cls.n_items, interaction_density=.5, num_user_features=200,
            num_item_features=200, n_features_per_user=20, n_features_per_item=20, pos_int_ratio=.5
        )

        cls.standard_model = TensorRec(n_components=10,
                                       n_tastes=3,
                                       user_repr_graph=NormalizedLinearRepresentationGraph(),
                                       attention_graph=LinearRepresentationGraph())
        cls.standard_model.fit(cls.interactions, cls.user_features, cls.item_features, epochs=10)

        cls.unbiased_model = TensorRec(n_components=10,
                                       n_tastes=3,
                                       biased=False,
                                       user_repr_graph=NormalizedLinearRepresentationGraph(),
                                       attention_graph=LinearRepresentationGraph())
        cls.unbiased_model.fit(cls.interactions, cls.user_features, cls.item_features, epochs=10)

    def test_predict_user_attn_repr(self):
        user_attn_repr = self.unbiased_model.predict_user_attention_representation(self.user_features)

        # attn repr should have shape [n_tastes, n_users, n_components]
        self.assertEqual(user_attn_repr.shape, (3, self.user_features.shape[0], 10))


class TensorRecAPIDatasetInputTestCase(TensorRecAPITestCase):

    @classmethod
    def setUpClass(cls):

        cls.n_users = 15
        cls.n_items = 30

        cls.interactions, cls.user_features, cls.item_features = generate_dummy_data(
            num_users=cls.n_users, num_items=cls.n_items, interaction_density=.5, num_user_features=200,
            num_item_features=200, n_features_per_user=20, n_features_per_item=20, pos_int_ratio=.5,
            return_datasets=True
        )

        cls.standard_model = TensorRec(n_components=10)
        cls.standard_model.fit(cls.interactions, cls.user_features, cls.item_features, epochs=10)

        cls.unbiased_model = TensorRec(n_components=10, biased=False)
        cls.unbiased_model.fit(cls.interactions, cls.user_features, cls.item_features, epochs=10)


class TensorRecAPITFRecordInputTestCase(TensorRecAPITestCase):

    @classmethod
    def setUpClass(cls):

        # Blow away an existing session to avoid 'tf_map_func not found' error
        set_session(None)

        cls.n_users = 15
        cls.n_items = 30

        int_ds, uf_ds, if_ds = generate_dummy_data(
            num_users=cls.n_users, num_items=cls.n_items, interaction_density=.5, num_user_features=200,
            num_item_features=200, n_features_per_user=20, n_features_per_item=20, pos_int_ratio=.5
        )

        cls.temp_dir = tempfile.mkdtemp()
        cls.interactions = os.path.join(cls.temp_dir, 'interactions.tfrecord')
        cls.user_features = os.path.join(cls.temp_dir, 'user_features.tfrecord')
        cls.item_features = os.path.join(cls.temp_dir, 'item_features.tfrecord')

        write_tfrecord_from_sparse_matrix(cls.interactions, int_ds)
        write_tfrecord_from_sparse_matrix(cls.user_features, uf_ds)
        write_tfrecord_from_sparse_matrix(cls.item_features, if_ds)

        cls.standard_model = TensorRec(n_components=10)
        cls.standard_model.fit(cls.interactions, cls.user_features, cls.item_features, epochs=10)

        cls.unbiased_model = TensorRec(n_components=10, biased=False)
        cls.unbiased_model.fit(cls.interactions, cls.user_features, cls.item_features, epochs=10)

    @classmethod
    def tearDownClass(cls):
        shutil.rmtree(cls.temp_dir)


class TensorRecSavingTestCase(TestCase):

    @classmethod
    def setUpClass(cls):
        cls.interactions, cls.user_features, cls.item_features = generate_dummy_data(
            num_users=15, num_items=30, interaction_density=.5, num_user_features=200, num_item_features=200,
            n_features_per_user=20, n_features_per_item=20, pos_int_ratio=.5
        )
        tf.reset_default_graph()
        set_session(None)

    def setUp(self):
        # Create a temporary directory
        self.test_dir = tempfile.mkdtemp()

    def tearDown(self):
        # Remove the directory after the test
        shutil.rmtree(self.test_dir)

    def test_save_and_load_model(self):
        model = TensorRec(n_components=10)
        model.fit(self.interactions, self.user_features, self.item_features, epochs=10)

        predictions = model.predict(user_features=self.user_features, item_features=self.item_features)
        ranks = model.predict_rank(user_features=self.user_features, item_features=self.item_features)
        model.save_model(directory_path=self.test_dir)

        # Check that, after saving, the same predictions come back
        predictions_after_save = model.predict(user_features=self.user_features, item_features=self.item_features)
        ranks_after_save = model.predict_rank(user_features=self.user_features, item_features=self.item_features)
        self.assertTrue((predictions == predictions_after_save).all())
        self.assertTrue((ranks == ranks_after_save).all())

        # Blow away the session
        set_session(None)
        tf.reset_default_graph()

        # Reload the model, predict, and check for equal predictions
        new_model = TensorRec.load_model(directory_path=self.test_dir)
        new_predictions = new_model.predict(user_features=self.user_features, item_features=self.item_features)
        new_ranks = new_model.predict_rank(user_features=self.user_features, item_features=self.item_features)

        self.assertTrue((predictions == new_predictions).all())
        self.assertTrue((ranks == new_ranks).all())

    def test_save_and_load_model_same_session(self):
        model = TensorRec(n_components=10)
        model.fit(self.interactions, self.user_features, self.item_features, epochs=10)

        predictions = model.predict(user_features=self.user_features, item_features=self.item_features)
        ranks = model.predict_rank(user_features=self.user_features, item_features=self.item_features)
        model.save_model(directory_path=self.test_dir)

        # Reload the model, predict, and check for equal predictions
        new_model = TensorRec.load_model(directory_path=self.test_dir)
        new_predictions = new_model.predict(user_features=self.user_features, item_features=self.item_features)
        new_ranks = new_model.predict_rank(user_features=self.user_features, item_features=self.item_features)

        self.assertTrue((predictions == new_predictions).all())
        self.assertTrue((ranks == new_ranks).all())



import abc
import tensorflow as tf


class AbstractRepresentationGraph(object):
    __metaclass__ = abc.ABCMeta

    @abc.abstractmethod
    def connect_representation_graph(self, tf_features, n_components, n_features, node_name_ending):
        """
        This method is responsible for connecting the user/item features to their respective latent representations.
        :param tf_features: tf.SparseTensor
        The user/item features as a SparseTensor of shape [ n_users, n_features ]
        :param n_components: int
        The size of the latent representation per user/item.
        :param n_features: int
        The size of the input features per user/item.
        :param node_name_ending: str
        A string, either 'user' or 'item', which can be added to TensorFlow node names for clarity.
        :return: tf.Tensor
        The user/item representation as a Tensor of shape [ n_users, n_components ]
        """
        pass


class LinearRepresentationGraph(AbstractRepresentationGraph):
    """
    Calculates the representation by passing the features through a linear embedding.
    Rough approximation of http://ceur-ws.org/Vol-1448/paper4.pdf
    """

    def connect_representation_graph(self, tf_features, n_components, n_features, node_name_ending):

        # Weights are normalized before building the variable
        tf.compat.v1.disable_eager_execution()
        raw_weights = tf.compat.v1.random_normal([n_features, n_components], stddev=1.0)
        normalized_weights = tf.nn.l2_normalize(raw_weights, 1)

        # Create variable nodes
        tf_linear_weights = tf.Variable(normalized_weights, name='linear_weights_{}'.format(node_name_ending))
        tf.compat.v1.disable_eager_execution()
        tf_repr = tf.compat.v1.sparse_tensor_dense_matmul(tf_features, tf_linear_weights)

        # Return repr layer and variables
        return tf_repr, [tf_linear_weights]


class NormalizedLinearRepresentationGraph(LinearRepresentationGraph):
    """
    Calculates the representation by passing the features through a linear embedding. Embeddings are L2 normalized,
    meaning all embeddings have equal magnitude. This can be useful as a user representation in mixture-of-tastes
    models, preventing one taste from having a much larger magnitude than others and dominating the recommendations.
    """

    def connect_representation_graph(self, tf_features, n_components, n_features, node_name_ending):
        tf_repr, weights_list = super(NormalizedLinearRepresentationGraph, self).connect_representation_graph(
            tf_features=tf_features, n_components=n_components, n_features=n_features, node_name_ending=node_name_ending
        )
        normalized_repr = tf.nn.l2_normalize(tf_repr, 1)
        return normalized_repr, weights_list


class FeaturePassThroughRepresentationGraph(AbstractRepresentationGraph):
    """
    Uses the features as the representation. This representation graph does no transformation to the features.
    """

    def connect_representation_graph(self, tf_features, n_components, n_features, node_name_ending):

        if n_components != n_features:
            raise ValueError('{} requires n_features and n_components to be equal. Either adjust n_components or use a '
                             'different representation graph. n_features = {}, n_components = {}'.format(
                                self.__class__.__name__, n_features, n_components
                             ))
        tf.compat.v1.disable_eager_execution()
        return tf.compat.v1.sparse_tensor_to_dense(tf_features, validate_indices=False), []


class WeightedFeaturePassThroughRepresentationGraph(FeaturePassThroughRepresentationGraph):
    """
    Uses the features as the representation. This representation graph learns weights for each feature.
    """

    def connect_representation_graph(self, tf_features, n_components, n_features, node_name_ending):
        dense_repr, _ = super(WeightedFeaturePassThroughRepresentationGraph, self).connect_representation_graph(
            tf_features=tf_features, n_components=n_components, n_features=n_features, node_name_ending=node_name_ending
        )

        weights = tf.ones([1, n_components])
        weighted_repr = tf.multiply(dense_repr, weights)
        return weighted_repr, [weights]


class ReLURepresentationGraph(AbstractRepresentationGraph):
    """
    Calculates the representations by passing the features through a single-layer ReLU neural network.
    :param relu_size: int or None
    The number of nodes in the ReLU layer. If None, the layer will be of size 4*n_components.
    """

    def __init__(self, relu_size=None):
        self.relu_size = relu_size

    def connect_representation_graph(self, tf_features, n_components, n_features, node_name_ending):

        # Infer ReLU layer size if necessary
        if self.relu_size is None:
            relu_size = 4 * n_components
        else:
            relu_size = self.relu_size

        # Create variable nodes
        tf_relu_weights = tf.Variable(tf.random_normal([n_features, relu_size], stddev=.5),
                                      name='relu_weights_{}'.format(node_name_ending))
        tf_relu_biases = tf.Variable(tf.zeros([1, relu_size]),
                                     name='relu_biases_{}'.format(node_name_ending))
        tf_linear_weights = tf.Variable(tf.random_normal([relu_size, n_components], stddev=.5),
                                        name='linear_weights_{}'.format(node_name_ending))

        # Create ReLU layer
        tf.compat.v1.disable_eager_execution()
        tf_relu = tf.nn.relu(tf.add(tf.compat.v1.sparse_tensor_dense_matmul(tf_features, tf_relu_weights),
                                    tf_relu_biases))
        tf_repr = tf.matmul(tf_relu, tf_linear_weights)

        # Return repr layer and variables
        return tf_repr, [tf_relu_weights, tf_linear_weights, tf_relu_biases]


class AbstractKerasRepresentationGraph(AbstractRepresentationGraph):
    """
    This abstract RepresentationGraph allows you to use Keras layers as a representation function by overriding the
    create_layers() method.
    """
    __metaclass__ = abc.ABCMeta

    def connect_representation_graph(self, tf_features, n_components, n_features, node_name_ending):
        layers = self.create_layers(n_features=n_features, n_components=n_components)

        weights = []
        last_layer = tf_features

        # Iterate through layers, connecting each one and extracting weights/biases
        for layer in layers:
            last_layer = layer(last_layer)
            if hasattr(layer, 'weights'):
                weights.extend(layer.weights)

        return last_layer, weights

    @abc.abstractmethod
    def create_layers(self, n_features, n_components):
        """
        Returns a list of Keras layers.
        :param n_features: int
        The input size of the first Keras layer.
        :param n_components: int
        The output size of the final Keras layer.
        :return: list
        A list of Keras layers.
        """
        pass



import numpy as np
import tensorflow as tf
import scipy.sparse as sp
from unittest import TestCase

from tensorrec.prediction_graphs import CosineSimilarityPredictionGraph
from tensorrec.recommendation_graphs import (
    project_biases, split_sparse_tensor_indices, bias_prediction_dense, bias_prediction_serial,
    densify_sampled_item_predictions, rank_predictions, collapse_mixture_of_tastes, predict_similar_items
)
from tensorrec.session_management import get_session


class RecommendationGraphsTestCase(TestCase):

    @classmethod
    def setUpClass(cls):
        cls.session = get_session()

    def test_project_biases(self):
        features = sp.coo_matrix([
            [1.0, 0, 0, 1.0],
            [0, 1.0, 0, 0],
            [0, 0, 1.0, 2.0],
        ], dtype=np.float32)
        n_features = 4

        sparse_tensor_features = tf.SparseTensor(np.mat([features.row, features.col]).transpose(),
                                                 features.data,
                                                 features.shape)
        tf_feature_biases, tf_projected_biases = project_biases(tf_features=sparse_tensor_features,
                                                                n_features=n_features)

        self.session.run(tf.global_variables_initializer())
        assign_op = tf_feature_biases.assign(value=[[-.5], [.5], [0], [2.0]])
        self.session.run(assign_op)

        result = tf_projected_biases.eval(session=self.session)
        expected_result = np.array([1.5, .5, 4.0])
        self.assertTrue((result == expected_result).all())

    def test_split_sparse_tensor_indices(self):
        interactions = sp.coo_matrix([
            [1.0, 0, 0, 1.0],
            [0, 1.0, 0, 0],
            [0, 0, 1.0, 2.0],
        ], dtype=np.float32)
        sparse_tensor_interactions = tf.SparseTensor(np.mat([interactions.row, interactions.col]).transpose(),
                                                     interactions.data,
                                                     interactions.shape)

        x_user, x_item = split_sparse_tensor_indices(tf_sparse_tensor=sparse_tensor_interactions, n_dimensions=2)

        x_user = x_user.eval(session=self.session)
        x_item = x_item.eval(session=self.session)

        expected_user = np.array([0, 0, 1, 2, 2])
        expected_item = np.array([0, 3, 1, 2, 3])

        self.assertTrue((x_user == expected_user).all())
        self.assertTrue((x_item == expected_item).all())

    def test_bias_prediction_dense(self):
        predictions = np.array([
            [1.0, 0, 0, 1.0],
            [0, 1.0, 0, 0],
            [0, 0, 1.0, 2.0],
        ], dtype=np.float32)

        projected_user_biases = np.array([0.0, 1.0, 2.0])
        projected_item_biases = np.array([0.0, -1.0, -2.0, -3.0])

        biased_predictions = bias_prediction_dense(
            tf_prediction=predictions,
            tf_projected_user_biases=projected_user_biases,
            tf_projected_item_biases=projected_item_biases
        ).eval(session=self.session)

        expected_biased_predictions = np.array([
            [1.0, -1.0, -2.0, -2.0],
            [1.0, 1.0, -1.0, -2.0],
            [2.0, 1.0, 1.0, 1.0],
        ], dtype=np.float32)

        self.assertTrue((biased_predictions == expected_biased_predictions).all())

    def test_bias_prediction_serial(self):
        predictions = np.array([1.0, 0, 0, 1.0, 0, 1.0, 0, 0, 0, 0, 1.0, 2.0], dtype=np.float32)

        projected_user_biases = np.array([0.0, 1.0, 2.0])
        projected_item_biases = np.array([0.0, -1.0, -2.0, -3.0])

        x_user = np.array([0, 0, 0, 0, 1, 1, 1, 1, 2, 2, 2, 2])
        x_item = np.array([0, 1, 2, 3, 0, 1, 2, 3, 0, 1, 2, 3])

        biased_predictions = bias_prediction_serial(
            tf_prediction_serial=predictions,
            tf_projected_user_biases=projected_user_biases,
            tf_projected_item_biases=projected_item_biases,
            tf_x_user=x_user,
            tf_x_item=x_item,
        ).eval(session=self.session)

        expected_biased_predictions = np.array([1.0, -1.0, -2.0, -2.0, 1.0, 1.0, -1.0, -2.0, 2.0, 1.0, 1.0, 1.0],
                                               dtype=np.float32)

        self.assertTrue((biased_predictions == expected_biased_predictions).all())

    def test_densify_sampled_item_predictions(self):
        input_data = np.array([1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12])
        result = densify_sampled_item_predictions(
            tf_sample_predictions_serial=input_data,
            tf_n_sampled_items=4,
            tf_n_users=3,
        ).eval(session=self.session)

        expected_result = np.array([
            [1, 2, 3, 4],
            [5, 6, 7, 8],
            [9, 10, 11, 12]
        ])
        self.assertTrue((result == expected_result).all())

    def test_rank_predictions(self):
        predictions = np.array([
            [1.0, 2.0, 3.0, 4.0],
            [4.0, 3.0, 2.0, 1.0],
            [3.0, 4.0, 1.0, 2.0],
        ], dtype=np.float32)

        ranked = rank_predictions(predictions).eval(session=self.session)

        expected_ranks = np.array([
            [4, 3, 2, 1],
            [1, 2, 3, 4],
            [2, 1, 4, 3],
        ], dtype=np.int)
        self.assertTrue((ranked == expected_ranks).all())

    def test_collapse_mixture_of_tastes(self):
        predictions = [
            np.array([1.0, 2.0, 3.0, 4.0], dtype=np.float32),
            np.array([4.0, 3.0, 2.0, 1.0], dtype=np.float32),
            np.array([3.0, 4.0, 1.0, 2.0], dtype=np.float32),
        ]

        collapsed_predictions = collapse_mixture_of_tastes(tastes_predictions=predictions,
                                                           tastes_attentions=None).eval(session=self.session)

        expected_predictions = np.array([[4.0, 4.0, 3.0, 4.0]], dtype=np.float32)
        self.assertTrue((collapsed_predictions == expected_predictions).all())

    def test_collapse_mixture_of_tastes_with_attention(self):
        predictions = [
            np.array([1.0, 2.0, 3.0, 4.0], dtype=np.float32),
            np.array([4.0, 3.0, 2.0, 1.0], dtype=np.float32),
            np.array([3.0, 4.0, 1.0, 2.0], dtype=np.float32),
        ]
        attentions = [
            np.array([1.0, 2.0, -3.0, 4.0], dtype=np.float32),
            np.array([0.0, 3.0, 2.0, 1.0], dtype=np.float32),
            np.array([3.0, 0.0, 1.0, 2.0], dtype=np.float32),
        ]

        collapsed_predictions = collapse_mixture_of_tastes(tastes_predictions=predictions,
                                                           tastes_attentions=attentions).eval(session=self.session)

        expected_predictions = np.array([[2.8136194, 2.7756228, 1.7372786, 3.6455793]], dtype=np.float32)
        self.assertTrue((collapsed_predictions == expected_predictions).all())

    def test_predict_similar_items(self):
        reprs = np.array([
            [1.0, 0.0, 0.0, 0.0],
            [0.0, 1.0, 0.0, 0.0],
            [0.0, -1.0, 0.0, 0.0],
        ], dtype=np.float32)
        sims = predict_similar_items(prediction_graph_factory=CosineSimilarityPredictionGraph(),
                                     tf_similar_items_ids=[1],
                                     tf_item_representation=reprs).eval(session=self.session)

        expected_sims = np.array([[0.0,  1.0, -1.0]], dtype=np.float32)
        self.assertTrue((sims == expected_sims).all())



from setuptools import setup

from os import path

here = path.abspath(path.dirname(__file__))

setup(
  name='kimrec',
  packages=['kimrec'],
  version='0.3',
  description='A TensorFlow recommendation algorithm and framework <Tensorrec> in Python modified by Hyeonjae Kim',
  author='Hyeonjae Kim',
  author_email='hyeonjae.gim@gmail.com',
  url='https://github.com/journey101/kimrec-for-tensorflow-recommendation',
  keywords=['machine-learning', 'tensorflow', 'recommendation-system', 'python', 'recommender-system'],
  classifiers=[],
  install_requires=[
      "numpy>=1.14.1",
      "scipy>=0.19.1",
      "six==1.11.0",
      "tensorflow>=2.4.1",
  ],
)



from nose_parameterized import parameterized
from unittest import TestCase

from tensorrec import TensorRec
from tensorrec.representation_graphs import (
    LinearRepresentationGraph, NormalizedLinearRepresentationGraph, FeaturePassThroughRepresentationGraph,
    WeightedFeaturePassThroughRepresentationGraph, ReLURepresentationGraph
)
from tensorrec.util import generate_dummy_data


class RepresentationGraphTestCase(TestCase):

    @parameterized.expand([
        ["linear", LinearRepresentationGraph, LinearRepresentationGraph, 50, 60, 20],
        ["norm_lin", NormalizedLinearRepresentationGraph, NormalizedLinearRepresentationGraph, 50, 60, 20],
        ["fpt_user", FeaturePassThroughRepresentationGraph, NormalizedLinearRepresentationGraph, 50, 60, 50],
        ["fpt_item", NormalizedLinearRepresentationGraph, FeaturePassThroughRepresentationGraph, 50, 60, 60],
        ["fpt_both", FeaturePassThroughRepresentationGraph, FeaturePassThroughRepresentationGraph, 50, 50, 50],
        ["weighted_fpt", WeightedFeaturePassThroughRepresentationGraph, WeightedFeaturePassThroughRepresentationGraph,
         50, 50, 50],
        ["relu", ReLURepresentationGraph, ReLURepresentationGraph, 50, 60, 20],
    ])
    def test_fit(self, name, user_repr, item_repr, n_user_features, n_item_features, n_components):
        interactions, user_features, item_features = generate_dummy_data(
            num_users=15, num_items=30, interaction_density=.5, num_user_features=n_user_features,
            num_item_features=n_item_features, n_features_per_user=20, n_features_per_item=20, pos_int_ratio=.5
        )
        model = TensorRec(n_components=n_components,
                          user_repr_graph=user_repr(),
                          item_repr_graph=item_repr())
        model.fit(interactions, user_features, item_features, epochs=10)

        # Ensure that the nodes have been built
        self.assertIsNotNone(model.tf_prediction)


class IdentityRepresentationGraphTestCase(TestCase):

    def test_fit_fail_on_bad_dims(self):
        interactions, user_features, item_features = generate_dummy_data(
            num_users=15, num_items=30, interaction_density=.5, num_user_features=30,
            num_item_features=20, n_features_per_user=20, n_features_per_item=20, pos_int_ratio=.5
        )

        with self.assertRaises(ValueError):
            model = TensorRec(n_components=25,
                              user_repr_graph=FeaturePassThroughRepresentationGraph(),
                              item_repr_graph=LinearRepresentationGraph())
            model.fit(interactions, user_features, item_features, epochs=10)

        with self.assertRaises(ValueError):
            model = TensorRec(n_components=25,
                              user_repr_graph=LinearRepresentationGraph(),
                              item_repr_graph=FeaturePassThroughRepresentationGraph())
            model.fit(interactions, user_features, item_features, epochs=10)
